from __future__ import annotations

import sys

from . import _import_wrapper

if not getattr(sys, 'frozen', False):
    __getattr__ = _import_wrapper.__getattr__

from ._version import __version__ as __version__



from __future__ import annotations

import enum
import typing
from collections.abc import Iterable
from functools import partial
from typing import Literal, Protocol, TypeVar

from attrs import asdict, evolve, frozen
from typing_extensions import Self
from yarl import URL

from .utils import StrEnum, fill

_TEnum = TypeVar('_TEnum', bound=enum.Enum)


class _FlavourKeyedEnum(Protocol[_TEnum]):
    Retail: _TEnum
    VanillaClassic: _TEnum
    Classic: _TEnum

    def __getitem__(self, __key: str) -> _TEnum:
        ...


class Flavour(StrEnum):
    # The latest Classic version is always aliased to "classic".
    # The logic here is that should Classic not be discontinued
    # it will continue to be updated in place so that new Classic versions
    # will inherit the "_classic_" folder.  This means we won't have to
    # migrate Classic profiles either automatically or by requiring user
    # intervention for new Classic releases.
    Retail = 'retail'
    VanillaClassic = 'vanilla_classic'
    Classic = 'classic'

    @classmethod
    def from_flavour_keyed_enum(cls, flavour_keyed_enum: enum.Enum) -> Self:
        return cls[flavour_keyed_enum.name]

    def to_flavour_keyed_enum(self, flavour_keyed_enum: _FlavourKeyedEnum[_TEnum]) -> _TEnum:
        return flavour_keyed_enum[self.name]


class FlavourVersionRange(enum.Enum):
    Retail = (
        range(1_00_00, 1_13_00), range(2_00_00, 2_05_00), range(3_00_00, 3_04_00), range(4_00_00, 11_00_00)  # fmt: skip
    )
    VanillaClassic = (range(1_13_00, 2_00_00),)
    Classic = (range(3_04_00, 4_00_00),)

    @classmethod
    def from_version_string(cls, version_string: str) -> Self | None:
        major, minor, patch = fill(map(int, version_string.split('.')), 0, 3)
        version_number = major * 1_00_00 + minor * 1_00 + patch
        return cls.from_version_number(version_number)

    @classmethod
    def from_version_number(cls, version_number: int) -> Self | None:
        return next((f for f in cls if f.is_within_version(version_number)), None)

    def is_within_version(self, version_number: int) -> bool:
        return any(version_number in r for r in self.value)


class Strategy(StrEnum):
    AnyFlavour = 'any_flavour'
    AnyReleaseType = 'any_release_type'
    VersionEq = 'version_eq'


class ChangelogFormat(StrEnum):
    Html = 'html'
    Markdown = 'markdown'
    Raw = 'raw'


@frozen
class SourceMetadata:
    id: str
    name: str
    strategies: frozenset[Strategy]
    changelog_format: ChangelogFormat
    addon_toc_key: typing.Union[str, None]


@frozen
class StrategyValues:
    any_flavour: Literal[True, None] = None
    any_release_type: Literal[True, None] = None
    version_eq: typing.Union[str, None] = None

    @property
    def filled_strategies(self) -> dict[Strategy, object]:
        return {Strategy(p): v for p, v in asdict(self).items() if v is not None}


_UNSOURCE = '*'
_UNSOURCE_NAME = 'ASTERISK'  # unicodedata.name(_UNSOURCE)

_STRATEGY_SEP = ','


@frozen(hash=True)
class Defn:
    source: str
    alias: str
    id: typing.Union[str, None] = None
    strategies: StrategyValues = StrategyValues()

    @classmethod
    def from_uri(
        cls,
        uri: str,
        *,
        known_sources: Iterable[str],
        allow_unsourced: bool,
        include_strategies: bool = False,
    ) -> Self:
        """Construct a ``Defn`` from a URI."""
        url = URL(uri)

        if url.scheme not in known_sources:
            if not allow_unsourced:
                raise ValueError(f'Unable to extract source from {uri}')
            elif url.scheme == _UNSOURCE:
                raise ValueError(f'{_UNSOURCE} ({_UNSOURCE_NAME}) is not valid as a source')

            make_cls = partial(cls, source=_UNSOURCE, alias=uri)
        else:
            make_cls = partial(cls, source=url.scheme, alias=url.path)

        if include_strategies:
            strategy_values = {
                s: v
                for f in url.fragment.split(_STRATEGY_SEP)
                for s, _, v in (f.partition('='),)
                if s
            }
            if strategy_values:
                unknown_strategies = strategy_values.keys() - set(Strategy)
                if unknown_strategies:
                    raise ValueError(f'Unknown strategies: {", ".join(unknown_strategies)}')

                make_cls = partial(
                    make_cls,
                    strategies=StrategyValues(
                        any_flavour='any_flavour' in strategy_values or None,
                        any_release_type='any_release_type' in strategy_values or None,
                        version_eq=strategy_values.get('version_eq'),
                    ),
                )

        return make_cls()

    def as_uri(self, include_strategies: bool = False) -> str:
        uri = f'{self.source}:{self.alias}'

        if include_strategies:
            filled_strategies = self.strategies.filled_strategies
            if filled_strategies:
                uri += f'''#{_STRATEGY_SEP.join(
                    s if v is True else f"{s}={v}" for s, v in filled_strategies.items()
                )}'''

        return uri

    def with_version(self, version: str) -> Self:
        return evolve(self, strategies=evolve(self.strategies, version_eq=version))

    @property
    def is_unsourced(self) -> bool:
        return self.source == _UNSOURCE


class AddonHashMethod(enum.Enum):
    Wowup = enum.auto()



from __future__ import annotations

import asyncio
import contextvars
import json
import os
import typing
from collections.abc import Awaitable, Callable, Iterator, Set
from contextlib import AsyncExitStack, contextmanager
from datetime import datetime
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Any, Literal, TypeVar

import aiohttp
import aiohttp.typedefs
import aiohttp.web
import anyio
import cattrs
import click
import iso8601
import sqlalchemy as sa
from aiohttp_rpc import JsonRpcMethod
from aiohttp_rpc import middlewares as rpc_middlewares
from aiohttp_rpc.errors import InvalidParams, ServerError
from aiohttp_rpc.server import WsJsonRpcServer
from attrs import evolve, frozen
from cattrs.preconf.json import configure_converter
from loguru import logger
from typing_extensions import Concatenate, ParamSpec, TypeAlias, TypedDict
from yarl import URL

from instawow import __version__, matchers, pkg_db, pkg_models
from instawow import results as R
from instawow._version import is_outdated
from instawow.cataloguer import ComputedCatalogueEntry
from instawow.common import Defn, Flavour, SourceMetadata
from instawow.config import Config, GlobalConfig, SecretStr, config_converter
from instawow.github_auth import get_codes, poll_for_access_token
from instawow.http import TraceRequestCtx, init_web_client
from instawow.manager import LocksType, Manager, bucketise_results, contextualise
from instawow.utils import WeakValueDefaultDictionary, read_resource_as_text, reveal_folder, uniq
from instawow.utils import run_in_thread as t
from instawow.wow_installations import infer_flavour_from_addon_dir

from . import frontend

_T = TypeVar('_T')
_P = ParamSpec('_P')
_ManagerBoundCoroFn: TypeAlias = Callable[Concatenate[Manager, _P], Awaitable[_T]]

_toga_handle = contextvars.ContextVar[tuple[Any, anyio.from_thread.BlockingPortal]]('_toga_handle')


LOCALHOST = '127.0.0.1'

_converter = cattrs.Converter(
    unstruct_collection_overrides={
        Set: sorted,
    }
)
configure_converter(_converter)
_converter.register_structure_hook(Path, lambda v, _: Path(v))
_converter.register_structure_hook(datetime, lambda v, _: iso8601.parse_date(v))
_converter.register_unstructure_hook(Path, str)


class _ConfigError(ServerError):
    code = -32001
    message = 'invalid configuration parameters'


class _ValidationErrorResponse(TypedDict):
    path: tuple[str | int, ...]
    message: str


def _transform_validation_errors(
    exc: cattrs.ClassValidationError | cattrs.IterableValidationError | BaseException,
    path: tuple[str | int, ...] = (),
) -> Iterator[_ValidationErrorResponse]:
    if isinstance(exc, cattrs.IterableValidationError):
        with_notes, _ = exc.group_exceptions()
        for exc, note in with_notes:
            new_path = (*path, note.index)
            if isinstance(exc, (cattrs.ClassValidationError, cattrs.IterableValidationError)):
                yield from _transform_validation_errors(exc, new_path)
            else:
                yield {
                    'path': new_path,
                    'message': str(exc),
                }
    elif isinstance(exc, cattrs.ClassValidationError):
        with_notes, _ = exc.group_exceptions()
        for exc, note in with_notes:
            new_path = (*path, note.name)
            if isinstance(exc, (cattrs.ClassValidationError, cattrs.IterableValidationError)):
                yield from _transform_validation_errors(exc, new_path)
            else:
                yield {
                    'path': new_path,
                    'message': str(exc),
                }


@contextmanager
def _reraise_validation_errors(
    error_class: type[ServerError | InvalidParams] = ServerError,
    values: dict[Any, Any] | None = None,
) -> Iterator[None]:
    try:
        yield
    except BaseException as exc:
        logger.info(f'invalid request: {(values, exc)}')
        raise error_class(data=list(_transform_validation_errors(exc))) from exc


@t
def _read_global_config() -> GlobalConfig:
    with _reraise_validation_errors(_ConfigError):
        return GlobalConfig.read()


@t
def _read_config(global_config: GlobalConfig, profile: str) -> Config:
    with _reraise_validation_errors(_ConfigError):
        return Config.read(global_config, profile)


_methods: list[tuple[str, type[BaseParams]]] = []


def _register_method(method: str):
    def inner(param_class: type[BaseParams]):
        _methods.append((method, param_class))
        return frozen(slots=False)(param_class)

    return inner


@frozen(slots=False)
class _ProfileParamMixin:
    profile: str


@frozen(slots=False)
class _DefnParamMixin:
    defns: list[Defn]


class BaseParams:
    @classmethod
    def bind(cls, method: str, managers: _ManagersManager) -> JsonRpcMethod:
        async def respond(**kwargs: Any):
            with _reraise_validation_errors(InvalidParams, kwargs):
                self = _converter.structure(kwargs, cls)
            return await self.respond(managers)

        return JsonRpcMethod(respond, name=method)

    async def respond(self, managers: _ManagersManager) -> Any:
        raise NotImplementedError


@_register_method('config/write_profile')
class WriteProfileConfigParams(_ProfileParamMixin, BaseParams):
    addon_dir: Path
    game_flavour: Flavour
    infer_game_flavour: bool

    async def respond(self, managers: _ManagersManager) -> Config:
        async with managers.locks['modify profile', self.profile]:
            with _reraise_validation_errors(_ConfigError):
                config = config_converter.structure(
                    {
                        'global_config': await _read_global_config(),
                        'profile': self.profile,
                        'addon_dir': self.addon_dir,
                        'game_flavour': self.game_flavour,
                    },
                    Config,
                )
                if self.infer_game_flavour:
                    config = evolve(
                        config,
                        game_flavour=infer_flavour_from_addon_dir(config.addon_dir)
                        or Flavour.Retail,
                    )
                await t(config.write)()

            # Unload the corresponding ``Manager`` instance for the
            # config to be reloaded on the next request
            managers.unload_profile(config.profile)

            return config


@_register_method('config/read_profile')
class ReadProfileConfigParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> Config:
        return await _read_config(managers.global_config, self.profile)


@_register_method('config/delete_profile')
class DeleteProfileConfigParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> None:
        async def delete_profile(manager: Manager):
            await t(manager.config.delete)()
            managers.unload_profile(self.profile)

        await managers.run(self.profile, delete_profile)


@_register_method('config/list_profiles')
class ListProfilesParams(BaseParams):
    async def respond(self, managers: _ManagersManager) -> list[str]:
        return await t(managers.global_config.list_profiles)()


@_register_method('config/update_global')
class UpdateGlobalConfigParams(BaseParams):
    access_tokens: dict[str, typing.Union[str, None]]

    async def respond(self, managers: _ManagersManager) -> GlobalConfig:
        def update_global_config_cb(global_config: GlobalConfig):
            return evolve(
                global_config,
                access_tokens=evolve(
                    global_config.access_tokens,
                    **{k: t if t is None else SecretStr(t) for k, t in self.access_tokens.items()},
                ),
            )

        return await managers.update_global_config(update_global_config_cb)


@_register_method('config/read_global')
class ReadGlobalConfigParams(BaseParams):
    async def respond(self, managers: _ManagersManager) -> GlobalConfig:
        return await _read_global_config()


class GithubCodesResponse(TypedDict):
    user_code: str
    verification_uri: str


@_register_method('config/initiate_github_auth_flow')
class InitiateGithubAuthFlowParams(BaseParams):
    async def respond(self, managers: _ManagersManager) -> GithubCodesResponse:
        return await managers.initiate_github_auth_flow()


class GithubAuthFlowStatusReport(TypedDict):
    status: Literal['success', 'failure']


@_register_method('config/query_github_auth_flow_status')
class QueryGithubAuthFlowStatusParams(BaseParams):
    async def respond(self, managers: _ManagersManager) -> GithubAuthFlowStatusReport:
        return {'status': await managers.wait_for_github_auth_completion()}


@_register_method('config/cancel_github_auth_flow')
class CancelGithubAuthFlowParams(BaseParams):
    async def respond(self, managers: _ManagersManager) -> None:
        managers.cancel_github_auth_polling()


@_register_method('sources/list')
class ListSourcesParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> dict[str, SourceMetadata]:
        manager = await managers.get_manager(self.profile)
        return {r.metadata.id: r.metadata for r in manager.resolvers.values()}


@_register_method('list')
class ListInstalledParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> list[pkg_models.Pkg]:
        manager = await managers.get_manager(self.profile)

        with manager.database.connect() as connection:
            installed_pkgs = (
                connection.execute(
                    sa.select(pkg_db.pkg).order_by(sa.func.lower(pkg_db.pkg.c.name))
                )
                .mappings()
                .all()
            )
            return [pkg_models.Pkg.from_row_mapping(connection, p) for p in installed_pkgs]


@_register_method('search')
class SearchParams(_ProfileParamMixin, BaseParams):
    search_terms: str
    limit: int
    sources: set[str]
    start_date: typing.Union[datetime, None]
    installed_only: bool

    async def respond(self, managers: _ManagersManager) -> list[ComputedCatalogueEntry]:
        return await managers.run(
            self.profile,
            partial(
                Manager.search,
                search_terms=self.search_terms,
                limit=self.limit,
                sources=self.sources,
                start_date=self.start_date,
                filter_installed='include_only' if self.installed_only else 'ident',
            ),
        )


class SuccessResult(TypedDict):
    status: Literal['success']
    addon: pkg_models.Pkg


class ErrorResult(TypedDict):
    status: Literal['failure', 'error']
    message: str


@_register_method('resolve')
class ResolveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def _resolve(self, manager: Manager):
        def extract_source(defn: Defn):
            if defn.is_unsourced:
                match = manager.pair_uri(defn.alias)
                if match:
                    source, alias = match
                    defn = evolve(defn, source=source, alias=alias)
            return defn

        return await manager.resolve(list(map(extract_source, self.defns)))

    async def respond(self, managers: _ManagersManager) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(self.profile, self._resolve)
        return [
            {'status': 'success', 'addon': r}
            if isinstance(r, pkg_models.Pkg)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('install')
class InstallParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    replace: bool

    async def respond(self, managers: _ManagersManager) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.install, defns=self.defns, replace=self.replace)
        )
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('update')
class UpdateParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.update, defns=self.defns, retain_defn_strategy=True)
        )
        return [
            {'status': r.status, 'addon': r.new_pkg}
            if isinstance(r, R.PkgUpdated)
            else {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('remove')
class RemoveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    keep_folders: bool

    async def respond(self, managers: _ManagersManager) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.remove, defns=self.defns, keep_folders=self.keep_folders)
        )
        return [
            {'status': r.status, 'addon': r.old_pkg}
            if isinstance(r, R.PkgRemoved)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('pin')
class PinParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(self.profile, partial(Manager.pin, defns=self.defns))
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('get_changelog')
class GetChangelogParams(_ProfileParamMixin, BaseParams):
    source: str
    changelog_url: str

    async def respond(self, managers: _ManagersManager) -> str:
        return await managers.run(
            self.profile,
            partial(Manager.get_changelog, source=self.source, uri=self.changelog_url),
        )


class AddonMatch(TypedDict):
    folders: list[AddonMatch_AddonFolder]
    matches: list[pkg_models.Pkg]


class AddonMatch_AddonFolder(TypedDict):
    name: str
    version: str


@_register_method('reconcile')
class ReconcileParams(_ProfileParamMixin, BaseParams):
    matcher: str

    async def respond(self, managers: _ManagersManager) -> list[AddonMatch]:
        manager = await managers.get_manager(self.profile)

        leftovers = await t(matchers.get_unreconciled_folders)(manager)

        match_groups = await matchers.DEFAULT_MATCHERS[self.matcher](manager, leftovers=leftovers)

        resolved_defns = await manager.resolve(uniq(d for _, b in match_groups for d in b))
        pkgs, _ = bucketise_results(resolved_defns.items())

        matched_folders = [
            (a, [i for i in (pkgs.get(d) for d in s) if i]) for a, s in match_groups
        ]
        unmatched_folders = (
            ([a], list[pkg_models.Pkg]())
            for a in sorted(leftovers - frozenset(i for a, _ in matched_folders for i in a))
        )
        return [
            {'folders': [{'name': f.name, 'version': f.version} for f in a], 'matches': m}
            for a, m in chain(matched_folders, unmatched_folders)
        ]


class ReconcileInstalledCandidate(TypedDict):
    installed_addon: pkg_models.Pkg
    alternative_addons: list[pkg_models.Pkg]


@_register_method('get_reconcile_installed_candidates')
class GetReconcileInstalledCandidatesParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> list[ReconcileInstalledCandidate]:
        manager = await managers.get_manager(self.profile)

        with manager.database.connect() as connection:
            installed_pkgs = [
                pkg_models.Pkg.from_row_mapping(connection, p)
                for p in connection.execute(
                    sa.select(pkg_db.pkg).order_by(sa.func.lower(pkg_db.pkg.c.name))
                )
                .mappings()
                .all()
            ]

        defn_groups = await managers.run(
            self.profile, partial(Manager.find_equivalent_pkg_defns, pkgs=installed_pkgs)
        )
        resolved_defns = await managers.run(
            self.profile,
            partial(Manager.resolve, defns=uniq(d for b in defn_groups.values() for d in b)),
        )
        pkgs, _ = bucketise_results(resolved_defns.items())
        return [
            {'installed_addon': p, 'alternative_addons': m}
            for p, s in defn_groups.items()
            for m in ([i for i in (pkgs.get(d) for d in s) if i],)
            if m
        ]


class DownloadProgressReport(TypedDict):
    defn: Defn
    progress: float


@_register_method('get_download_progress')
class GetDownloadProgressParams(_ProfileParamMixin, BaseParams):
    async def respond(self, managers: _ManagersManager) -> list[DownloadProgressReport]:
        return [
            {'defn': p.to_defn(), 'progress': r}
            for p, r in await managers.get_manager_download_progress(self.profile)
        ]


class GetVersionResult(TypedDict):
    installed_version: str
    new_version: str | None


@_register_method('meta/get_version')
class GetVersionParams(BaseParams):
    async def respond(self, managers: _ManagersManager) -> GetVersionResult:
        outdated, new_version = await is_outdated(managers.global_config)
        return {
            'installed_version': __version__,
            'new_version': new_version if outdated else None,
        }


@_register_method('assist/open_url')
class OpenUrlParams(BaseParams):
    url: str

    async def respond(self, managers: _ManagersManager) -> None:
        click.launch(self.url)


@_register_method('assist/reveal_folder')
class RevealFolderParams(BaseParams):
    path_parts: list[str]

    async def respond(self, managers: _ManagersManager) -> None:
        reveal_folder(os.path.join(*self.path_parts))


class SelectFolderResult(TypedDict):
    selection: Path | None


@_register_method('assist/select_folder')
class SelectFolderParams(BaseParams):
    initial_folder: typing.Union[str, None]

    async def respond(self, managers: _ManagersManager) -> SelectFolderResult:
        main_window, portal = _toga_handle.get()

        async def select_folder() -> Path | None:
            return await main_window.select_folder_dialog('Select folder', self.initial_folder)

        try:
            selection = portal.start_task_soon(select_folder).result()
        except ValueError:
            selection = None
        return {'selection': selection}


class ConfirmDialogueResult(TypedDict):
    ok: bool


@_register_method('assist/confirm')
class ConfirmDialogueParams(BaseParams):
    title: str
    message: str

    async def respond(self, managers: _ManagersManager) -> ConfirmDialogueResult:
        main_window, portal = _toga_handle.get()

        async def confirm() -> bool:
            return await main_window.confirm_dialog(self.title, self.message)

        return {'ok': portal.start_task_soon(confirm).result()}


def _init_json_rpc_web_client(cache_dir: Path):
    async def do_on_request_end(
        client_session: aiohttp.ClientSession,
        trace_config_ctx: Any,
        params: aiohttp.TraceRequestEndParams,
    ) -> None:
        trace_request_ctx: TraceRequestCtx = trace_config_ctx.trace_request_ctx
        if (
            trace_request_ctx
            and trace_request_ctx['report_progress'] == 'pkg_download'
            and params.response.content_length
            and aiohttp.hdrs.CONTENT_ENCODING not in params.response.headers
        ):
            content = params.response.content
            content_length = params.response.content_length
            entry = (
                trace_request_ctx['profile'],
                trace_request_ctx['pkg'],
                lambda: content.total_bytes / content_length,
            )
            progress_reporters.add(entry)
            content.on_eof(lambda: progress_reporters.remove(entry))

    progress_reporters: set[tuple[str, pkg_models.Pkg, Callable[[], float]]] = set()

    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()

    return (init_web_client(cache_dir, trace_configs=[trace_config]), progress_reporters)


class _ManagersManager:
    def __init__(self):
        self._exit_stack = AsyncExitStack()

        self.locks: LocksType = WeakValueDefaultDictionary(asyncio.Lock)

        self._managers = dict[str, Manager]()

        self._github_auth_device_codes = None
        self._github_auth_flow_task = None

    async def __aenter__(self):
        self.global_config = await _read_global_config()
        init_json_rpc_web_client, self._download_progress_reporters = _init_json_rpc_web_client(
            self.global_config.cache_dir
        )
        self._web_client = await self._exit_stack.enter_async_context(init_json_rpc_web_client)
        contextualise(web_client=self._web_client, locks=self.locks)

    async def __aexit__(self, *args: object):
        self.cancel_github_auth_polling()
        await self._exit_stack.aclose()

    def unload_profile(self, profile: str):
        if profile in self._managers:
            del self._managers[profile]

    def _unload_all_profiles(self):
        self._managers.clear()

    async def update_global_config(
        self, update_cb: Callable[[GlobalConfig], GlobalConfig]
    ) -> GlobalConfig:
        async with self.locks['update global config']:
            with _reraise_validation_errors(_ConfigError):
                self.global_config = update_cb(self.global_config)
                await t(self.global_config.write)()

            self._unload_all_profiles()

            return self.global_config

    async def run(self, profile: str, coro_fn: _ManagerBoundCoroFn[..., _T]) -> _T:
        try:
            manager = self._managers[profile]
        except KeyError:
            async with self.locks['modify profile', profile]:
                try:
                    manager = self._managers[profile]
                except KeyError:
                    manager = self._managers[profile] = Manager.from_config(
                        await _read_config(self.global_config, profile)
                    )

        return await coro_fn(manager)

    async def get_manager(self, profile: str):
        async def get_manager(manager: Manager):
            return manager

        return await self.run(profile, get_manager)

    async def get_manager_download_progress(self, profile: str):
        manager = await self.get_manager(profile)
        return (
            (p, r())
            for m, p, r in self._download_progress_reporters
            if m == manager.config.profile
        )

    async def initiate_github_auth_flow(self):
        async with self.locks['initiate github auth flow']:
            if self._github_auth_device_codes is None:

                async def finalise_github_auth_flow():
                    result = await poll_for_access_token(
                        self._web_client, codes['device_code'], codes['interval']
                    )

                    def update_global_config_cb(global_config: GlobalConfig):
                        return evolve(
                            global_config,
                            access_tokens=evolve(
                                global_config.access_tokens, github=SecretStr(result)
                            ),
                        )

                    await self.update_global_config(update_global_config_cb)

                def on_task_complete(task: object):
                    self._github_auth_flow_task = None
                    self._github_auth_device_codes = None

                self._github_auth_device_codes = codes = await get_codes(self._web_client)
                self._github_auth_flow_task = asyncio.create_task(finalise_github_auth_flow())
                self._github_auth_flow_task.add_done_callback(on_task_complete)

        return self._github_auth_device_codes

    async def wait_for_github_auth_completion(self):
        if self._github_auth_flow_task is not None:
            try:
                await self._github_auth_flow_task
            except BaseException:
                return 'failure'
        return 'success'

    def cancel_github_auth_polling(self):
        if self._github_auth_flow_task is not None:
            self._github_auth_flow_task.cancel()


async def create_app(toga_handle: tuple[Any, anyio.from_thread.BlockingPortal] | None = None):
    if toga_handle:
        _toga_handle.set(toga_handle)

    managers = _ManagersManager()

    async def managers_listen(app: aiohttp.web.Application):
        async with managers:
            yield

    async def get_index(request: aiohttp.web.Request):
        return aiohttp.web.Response(
            content_type='text/html',
            text=read_resource_as_text(frontend, 'index.html'),
        )

    async def get_static_file(request: aiohttp.web.Request):
        filename = request.path.lstrip('/')
        if filename.startswith('svelte-bundle.js'):
            content_type = 'application/javascript'
        elif filename == 'svelte-bundle.css':
            content_type = 'text/css'
        else:
            raise aiohttp.web.HTTPNotFound

        return aiohttp.web.Response(
            content_type=content_type,
            text=read_resource_as_text(frontend, filename),
        )

    def json_serialize(value: dict[str, Any]):
        return json.dumps(_converter.unstructure(value))

    rpc_server = WsJsonRpcServer(
        json_serialize=json_serialize,
        middlewares=rpc_middlewares.DEFAULT_MIDDLEWARES,  # pyright: ignore[reportUnknownMemberType]
    )
    rpc_server.add_methods(  # pyright: ignore[reportUnknownMemberType]
        [m.bind(n, managers) for n, m in _methods]
    )

    @aiohttp.web.middleware
    async def enforce_same_origin(
        request: aiohttp.web.Request,
        handler: aiohttp.typedefs.Handler,
    ):
        if request.remote != LOCALHOST:
            raise aiohttp.web.HTTPUnauthorized

        if request.path.startswith('/api'):
            origin = request.headers.get(aiohttp.hdrs.ORIGIN)
            if origin is None:
                raise aiohttp.web.HTTPUnauthorized

            origin_url = URL(origin)
            if (
                origin_url.scheme != request.url.scheme
                or origin_url.host != request.url.host
                or origin_url.port != request.url.port
            ):
                raise aiohttp.web.HTTPUnauthorized

        return await handler(request)

    app = aiohttp.web.Application(middlewares=[enforce_same_origin])
    app.add_routes(
        [
            aiohttp.web.get('/', get_index),
            aiohttp.web.get(
                r'/svelte-bundle{extension:(?:\.css|\.js(?:\.map)?)}', get_static_file
            ),
            aiohttp.web.get('/api', rpc_server.handle_http_request),
        ]
    )
    app.cleanup_ctx.append(managers_listen)
    app.freeze()
    return app


async def run_app(app: aiohttp.web.Application):
    "Fire up the server."
    app_runner = aiohttp.web.AppRunner(app)
    await app_runner.setup()
    assert app_runner.server  # Server is created during setup
    # By omitting the port ``loop.create_server`` will find an available port
    # to bind to - equivalent to creating a socket on port 0.
    server = await asyncio.get_running_loop().create_server(app_runner.server, LOCALHOST)
    host, port = server.sockets[0].getsockname()

    async def serve():
        try:
            # ``server_forever`` cleans up after the server when it's interrupted
            await server.serve_forever()
        finally:
            await app_runner.cleanup()

    return (
        URL.build(scheme='http', host=host, port=port),
        serve,
    )



from __future__ import annotations

import asyncio
from collections.abc import Callable
from pathlib import Path
from typing import Any

import aiohttp
import aiohttp.web
import pytest
from aresponses import ResponsesMockServer
from aresponses.errors import NoRouteFoundError
from loguru import logger

from instawow.common import Flavour
from instawow.config import Config, GlobalConfig
from instawow.http import init_web_client
from instawow.manager import Manager, contextualise

from .fixtures.http import ROUTES


def pytest_addoption(parser: pytest.Parser):
    parser.addoption('--iw-no-mock-http', action='store_true')


def should_mock(fn: Callable[..., object]):
    import inspect

    def wrapper(request: pytest.FixtureRequest):
        if request.config.getoption('--iw-no-mock-http'):
            return None
        elif any(m.name == 'iw_no_mock_http' for m in request.node.iter_markers()):
            return None

        args = (request.getfixturevalue(p) for p in inspect.signature(fn).parameters)
        return fn(*args)

    return wrapper


@pytest.fixture
def caplog(caplog: pytest.LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format='{message}',
        level=0,
        filter=lambda record: record['level'].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)


class _StrictResponsesMockServer(ResponsesMockServer):
    async def _find_response(self, request: aiohttp.web.Request):
        response = await super()._find_response(request)
        if response == (None, None):
            raise NoRouteFoundError(f'No match found for <{request.method} {request.url}>')
        return response


@pytest.fixture
async def aresponses(event_loop: asyncio.AbstractEventLoop):
    async with _StrictResponsesMockServer(loop=event_loop) as server:
        yield server


@pytest.fixture(params=['foo'])
def iw_global_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    return {
        'temp_dir': tmp_path / 'temp',
        'config_dir': tmp_path / 'config',
        'access_tokens': {
            'cfcore': request.param,
            'github': None,
            'wago': None,
            'wago_addons': request.param,
        },
    }


@pytest.fixture(params=[Flavour.Retail])
def iw_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    addons = tmp_path / 'wow' / 'interface' / 'addons'
    addons.mkdir(parents=True)
    return {'profile': '__default__', 'addon_dir': addons, 'game_flavour': request.param}


@pytest.fixture
def iw_config(iw_config_values: dict[str, Any], iw_global_config_values: dict[str, Any]):
    global_config = GlobalConfig.from_values(iw_global_config_values).write()
    return Config.from_values({'global_config': global_config, **iw_config_values}).write()


@pytest.fixture(autouse=True)
async def _iw_global_config_defaults(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_global_config_values['config_dir']))
    monkeypatch.setenv('INSTAWOW_TEMP_DIR', str(iw_global_config_values['temp_dir']))


@pytest.fixture
async def iw_web_client(iw_config: Config):
    async with init_web_client(iw_config.global_config.cache_dir) as web_client:
        yield web_client


@pytest.fixture
def iw_manager(iw_config: Config, iw_web_client: aiohttp.ClientSession):
    contextualise(web_client=iw_web_client)
    return Manager.from_config(iw_config)


@pytest.fixture(autouse=True, params=['all'])
@should_mock
def _iw_mock_aiohttp_requests(
    request: pytest.FixtureRequest, aresponses: _StrictResponsesMockServer
):
    if request.param == 'all':
        routes = ROUTES.values()
    else:
        if not request.param.issubset(ROUTES.keys()):
            raise ValueError('Supplied routes must be subset of all routes')

        routes = (ROUTES[k] for k in ROUTES.keys() & request.param)

    for route in routes:
        aresponses.add(**route.to_aresponses_add_args())



from __future__ import annotations

from collections.abc import AsyncIterator
from datetime import datetime, timedelta, timezone

from loguru import logger
from typing_extensions import TypedDict
from yarl import URL

from .. import pkg_models
from .. import results as R
from ..cataloguer import CatalogueEntry
from ..common import ChangelogFormat, Defn, Flavour, FlavourVersionRange, SourceMetadata
from ..http import ClientSessionType
from ..resolvers import BaseResolver


class _TukuiAddon(TypedDict):
    id: int
    slug: str
    author: str
    name: str
    url: str
    version: str
    changelog_url: str
    ticket_url: str
    git_url: str
    patch: list[str]
    last_update: str  # YYYY-MM-DD
    web_url: str
    donate_url: str
    small_desc: str
    screenshort_url: str
    directories: list[str]


class TukuiResolver(BaseResolver):
    metadata = SourceMetadata(
        id='tukui',
        name='Tukui',
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key='X-Tukui-ProjectID',
    )
    requires_access_token = None

    _api_url = URL('https://api.tukui.org/v1/')

    async def resolve_one(self, defn: Defn, metadata: None) -> pkg_models.Pkg:
        async with self._manager.web_client.get(
            self._api_url / 'addon' / defn.alias,
            expire_after=timedelta(minutes=5),
        ) as response:
            if response.status == 404:
                raise R.PkgNonexistent
            response.raise_for_status()

            ui_metadata: _TukuiAddon = await response.json()

        if not any(
            Flavour.from_flavour_keyed_enum(r) is self._manager.config.game_flavour
            for g in ui_metadata['patch']
            for r in (FlavourVersionRange.from_version_string(g),)
            if r
        ):
            raise R.PkgFilesNotMatching(defn.strategies)

        return pkg_models.Pkg(
            source=self.metadata.id,
            id=str(ui_metadata['id']),
            slug=ui_metadata['slug'],
            name=ui_metadata['name'],
            description=ui_metadata['small_desc'],
            url=ui_metadata['web_url'],
            download_url=ui_metadata['url'],
            date_published=datetime.fromisoformat(ui_metadata['last_update']).replace(
                tzinfo=timezone.utc
            ),
            version=ui_metadata['version'],
            changelog_url=(
                # The changelog URL is not versioned - add fragment to allow caching.
                str(URL(ui_metadata['changelog_url']).with_fragment(ui_metadata['version']))
            ),
            options=pkg_models.PkgOptions.from_strategy_values(defn.strategies),
        )

    @classmethod
    async def catalogue(cls, web_client: ClientSessionType) -> AsyncIterator[CatalogueEntry]:
        url = cls._api_url / 'addons'
        logger.debug(f'retrieving {url}')

        async with web_client.get(url, raise_for_status=True) as response:
            items: list[_TukuiAddon] = await response.json()

        for item in items:
            yield CatalogueEntry(
                source=cls.metadata.id,
                id=str(item['id']),
                slug=item['slug'],
                name=item['name'],
                url=item['web_url'],
                game_flavours=frozenset(
                    Flavour.from_flavour_keyed_enum(r)
                    for g in item['patch']
                    for r in (FlavourVersionRange.from_version_string(g),)
                    if r
                ),
                download_count=1,
                last_updated=datetime.fromisoformat(item['last_update']).replace(
                    tzinfo=timezone.utc
                ),
                folders=[frozenset(item['directories'])],
            )



from __future__ import annotations

from enum import Enum

from instawow.common import Flavour, FlavourVersionRange


def test_can_convert_between_flavour_keyed_enum_and_flavour():
    class Foo(Enum):
        Retail = 1

    assert Flavour.from_flavour_keyed_enum(Foo.Retail) is Flavour.Retail
    assert Flavour.Retail.to_flavour_keyed_enum(Foo) is Foo.Retail


def test_can_extract_flavour_from_version_number():
    assert FlavourVersionRange.from_version_number(95000) is FlavourVersionRange.Retail
    assert FlavourVersionRange.from_version_number(34000) is FlavourVersionRange.Classic
    assert FlavourVersionRange.from_version_number(12300) is FlavourVersionRange.VanillaClassic


def test_can_extract_flavour_from_version_string():
    assert FlavourVersionRange.from_version_string('9.50.0') is FlavourVersionRange.Retail
    assert FlavourVersionRange.from_version_string('3.40.0') is FlavourVersionRange.Classic
    assert FlavourVersionRange.from_version_string('1.23.0') is FlavourVersionRange.VanillaClassic


def test_can_extract_flavour_from_partial_version_string():
    assert FlavourVersionRange.from_version_string('9.2') is FlavourVersionRange.Retail
    assert FlavourVersionRange.from_version_string('3.4') is FlavourVersionRange.Classic
    assert FlavourVersionRange.from_version_string('3') is FlavourVersionRange.Retail



from __future__ import annotations

from collections.abc import Collection, Iterable
from datetime import timedelta
from typing import Literal, NoReturn

import iso8601
from typing_extensions import TypedDict
from yarl import URL

from .. import pkg_models
from .. import results as R
from ..common import AddonHashMethod, ChangelogFormat, Defn, SourceMetadata, Strategy
from ..http import make_generic_progress_ctx
from ..resolvers import BaseResolver, HeadersIntent, TFolderHashCandidate
from ..utils import StrEnum, as_plain_text_data_url, run_in_thread

_WagoStability = Literal['stable', 'beta', 'alpha']


class _WagoGameVersion(StrEnum):
    Retail = 'retail'
    VanillaClassic = 'classic'
    Classic = 'wotlk'


class _WagoMatchRequest(TypedDict):
    "``/addons/_match``"
    game_version: _WagoGameVersion
    addons: list[_WagoMatchRequestAddon]


class _WagoMatchRequestAddon(TypedDict):
    name: str
    hash: str


class _WagoMatches(TypedDict):
    "``/addons/_match``"
    addons: list[_WagoMatchingAddon | None]


class _WagoMatchingAddon(TypedDict):
    id: str
    name: str  # Eq to ``WagoAddon.display_name``
    authors: list[str]
    website_url: str  # Page on Wago
    thumbnail: str | None
    matched_release: _WagoRecentAddonRelease
    modules: dict[str, _WagoAddonModule]  # Add-on folders
    cf: str | None
    wowi: str | None
    wago: str  # Same as ``id``
    recent_releases: dict[_WagoStability, _WagoRecentAddonRelease]


class _WagoRecentAddonRelease(TypedDict):
    id: str
    label: str  # Version
    patch: str  # e.g. "9.2.5"
    created_at: str  # ISO date-time
    link: str


class _WagoAddonModule(TypedDict):
    hash: str


class _WagoAddon(TypedDict):
    "``/addons/{id}``"
    id: str
    slug: str
    display_name: str  # Eq to ``WagoRecentAddon.name``
    thumbnail_image: str | None  # Eq to ``WagoRecentAddon.thumbnail``
    summary: str
    description: str  # Long description
    website: str  # Author website
    gallery: list[str]
    authors: list[str]
    download_count: int
    website_url: str  # Page on Wago
    recent_release: dict[_WagoStability, _WagoAddonRelease] | list[NoReturn]


class _WagoAddonRelease(TypedDict):
    label: str  # Version
    supported_retail_patch: str | None  # e.g. "9.2.5"
    supported_classic_patch: str | None
    supported_wotlk_patch: str | None
    changelog: str
    stability: _WagoStability
    created_at: str  # ISO date-time
    download_link: str


class WagoResolver(BaseResolver):
    metadata = SourceMetadata(
        id='wago',
        name='Wago Addons',
        strategies=frozenset({Strategy.AnyReleaseType}),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key='X-Wago-ID',
    )
    requires_access_token = 'wago_addons'

    _wago_external_api_url = URL('https://addons.wago.io/api/external')

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        if url.host == 'addons.wago.io' and len(url.parts) > 2 and url.parts[1] == 'addons':
            return url.parts[2]

    async def make_request_headers(self, intent: HeadersIntent | None = None) -> dict[str, str]:
        maybe_access_token = self._get_access_token(self._manager.config.global_config)
        if maybe_access_token is None:
            raise ValueError(f'{self.metadata.name} access token is not configured')
        return {'Authorization': f'Bearer {maybe_access_token}'}

    async def resolve_one(self, defn: Defn, metadata: None) -> pkg_models.Pkg:
        wago_game_version = self._manager.config.game_flavour.to_flavour_keyed_enum(
            _WagoGameVersion
        )

        async with self._manager.web_client.get(
            (self._wago_external_api_url / 'addons' / defn.alias).with_query(
                game_version=wago_game_version.value,
            ),
            expire_after=timedelta(minutes=5),
            headers=await self.make_request_headers(),
        ) as response:
            if response.status == 404:
                raise R.PkgNonexistent
            response.raise_for_status()

            addon_metadata: _WagoAddon = await response.json()

        recent_releases = dict(addon_metadata['recent_release'])
        if not defn.strategies.any_release_type:
            files = filter(None, (recent_releases.get('stable'),))
        else:
            files = recent_releases.values()

        matching_file = max(
            ((iso8601.parse_date(f['created_at']), f) for f in files), default=None
        )
        if matching_file is None:
            raise R.PkgFilesMissing
        else:
            file_date, file = matching_file

        return pkg_models.Pkg(
            source=self.metadata.id,
            id=addon_metadata['id'],
            slug=addon_metadata['slug'],
            name=addon_metadata['display_name'],
            description=addon_metadata['summary'],
            url=addon_metadata['website_url'],
            download_url=file['download_link'],
            date_published=file_date,
            version=file['label'],
            changelog_url=as_plain_text_data_url(file['changelog']),
            options=pkg_models.PkgOptions.from_strategy_values(defn.strategies),
        )

    @run_in_thread
    def _make_match_params(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> _WagoMatchRequest:
        return {
            'game_version': self._manager.config.game_flavour.to_flavour_keyed_enum(
                _WagoGameVersion
            ),
            'addons': [
                {
                    'name': c.name,
                    'hash': c.hash_contents(AddonHashMethod.Wowup),
                }
                for c in candidates
            ],
        }

    async def get_folder_hash_matches(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> Iterable[tuple[Defn, frozenset[TFolderHashCandidate]]]:
        async with self._manager.web_client.post(
            self._wago_external_api_url / 'addons/_match',
            expire_after=timedelta(minutes=15),
            headers=await self.make_request_headers(),
            json=await self._make_match_params(candidates),
            raise_for_status=True,
            trace_request_ctx=make_generic_progress_ctx('Finding matching Wago add-ons'),
        ) as response:
            matches: _WagoMatches = await response.json()

        candidates_by_name = {c.name: c for c in candidates}

        return [
            (
                Defn(self.metadata.id, a['id']),
                # We are filtering out add-ons without same-flavour TOCs on our end,
                # so the API might return modules which aren't in the candidate list.
                frozenset(f for m in a['modules'] for f in (candidates_by_name.get(m),) if f),
            )
            for a in matches['addons']
            if a
        ]



from __future__ import annotations

import datetime
from pathlib import Path

import pytest
from aiohttp import ClientError
from attrs import evolve

from instawow import results as R
from instawow.common import Defn, Flavour, Strategy
from instawow.manager import Manager
from instawow.pkg_models import Pkg


def test_auth_bound_resolvers_are_not_unloaded_if_tokens_set(iw_manager: Manager):
    assert {
        r.metadata.id for r in iw_manager.RESOLVERS if r.requires_access_token is not None
    }.issubset(iw_manager.resolvers)


@pytest.mark.parametrize(
    'iw_global_config_values',
    [None],
    indirect=True,
)
def test_auth_bound_resolvers_are_unloaded_if_tokens_unset(iw_manager: Manager):
    assert {
        r.metadata.id for r in iw_manager.RESOLVERS if r.requires_access_token is not None
    }.isdisjoint(iw_manager.resolvers)


async def test_pinning_supported_pkg(iw_manager: Manager):
    defn = Defn('curse', 'molinari')

    install_result = (await iw_manager.install([defn], False))[defn]
    assert type(install_result) is R.PkgInstalled

    for new_defn in (defn.with_version(install_result.pkg.version), defn):
        pin_result = (await iw_manager.pin([new_defn]))[new_defn]
        assert type(pin_result) is R.PkgInstalled
        assert pin_result.pkg.options.version_eq is bool(new_defn.strategies.version_eq)
        assert install_result.pkg.version == pin_result.pkg.version


async def test_pinning_unsupported_pkg(iw_manager: Manager):
    molinari_defn = Defn('wowi', '13188')

    await iw_manager.install([molinari_defn], False)
    installed_pkg = iw_manager.get_pkg(molinari_defn)
    assert installed_pkg is not None
    assert installed_pkg.options.version_eq is False

    result = (await iw_manager.pin([molinari_defn]))[molinari_defn]
    assert type(result) is R.PkgStrategiesUnsupported
    assert result.strategies == {Strategy.VersionEq}
    assert installed_pkg.options.version_eq is False


async def test_pinning_nonexistent_pkg(iw_manager: Manager):
    molinari_defn = Defn('curse', 'molinari')
    result = await iw_manager.pin([molinari_defn])
    assert type(result[molinari_defn]) is R.PkgNotInstalled


async def test_pinning_unsupported_nonexistent_pkg(iw_manager: Manager):
    molinari_defn = Defn('wowi', '13188')
    result = await iw_manager.pin([molinari_defn])
    assert type(result[molinari_defn]) is R.PkgStrategiesUnsupported


@pytest.mark.parametrize('exception', [ValueError('foo'), ClientError('bar')])
async def test_resolve_rewraps_exception_appropriately_from_resolve(
    monkeypatch: pytest.MonkeyPatch, iw_manager: Manager, exception: Exception
):
    async def resolve_one(self, defn, metadata):
        raise exception

    monkeypatch.setattr('instawow._sources.cfcore.CfCoreResolver.resolve_one', resolve_one)

    defn = Defn('curse', 'molinari')
    result = (await iw_manager.resolve([defn]))[defn]
    assert type(result) is R.InternalError
    assert result.message == f'internal error: "{exception}"'


@pytest.mark.parametrize('exception', [ValueError('foo'), ClientError('bar')])
async def test_resolve_rewraps_exception_appropriately_from_batch_resolve(
    monkeypatch: pytest.MonkeyPatch, iw_manager: Manager, exception: Exception
):
    async def resolve(self, defns):
        raise exception

    monkeypatch.setattr('instawow._sources.cfcore.CfCoreResolver.resolve', resolve)

    defn = Defn('curse', 'molinari')
    result = (await iw_manager.resolve([defn]))[defn]
    assert type(result) is R.InternalError
    assert result.message == f'internal error: "{exception}"'


async def test_resolve_invalid_source(iw_manager: Manager):
    defn = Defn('bar', 'baz')
    results = await iw_manager.resolve([defn])
    assert type(results[defn]) is R.PkgSourceInvalid


async def test_resolve_plugin_hook_source(iw_manager: Manager):
    pytest.importorskip('instawow_test_plugin')
    defn = Defn('me', 'bar')
    results = await iw_manager.resolve([defn])
    assert type(results[defn]) is Pkg


async def test_install_can_replace_unreconciled_folders(iw_manager: Manager):
    molinari = iw_manager.config.addon_dir / 'Molinari'
    molinari.mkdir()

    defn = Defn('curse', 'molinari')

    result = await iw_manager.install([defn], replace=False)
    assert type(result[defn]) is R.PkgConflictsWithUnreconciled
    assert not any(molinari.iterdir())

    result = await iw_manager.install([defn], replace=True)
    assert type(result[defn]) is R.PkgInstalled
    assert any(molinari.iterdir())


async def test_install_cannot_replace_reconciled_folders(iw_manager: Manager):
    curse_defn = Defn('curse', 'molinari')
    wowi_defn = Defn('wowi', '13188-molinari')

    result = await iw_manager.install([curse_defn], replace=False)
    assert type(result[curse_defn]) is R.PkgInstalled

    result = await iw_manager.install([wowi_defn], replace=False)
    assert type(result[wowi_defn]) is R.PkgConflictsWithInstalled

    result = await iw_manager.install([wowi_defn], replace=True)
    assert type(result[wowi_defn]) is R.PkgConflictsWithInstalled


async def test_install_recognises_renamed_pkg_from_id(
    monkeypatch: pytest.MonkeyPatch, iw_manager: Manager
):
    old_defn = Defn('curse', 'molinari')
    new_defn = Defn('curse', 'molinarifico')

    result = await iw_manager.install([old_defn], replace=False)
    assert type(result[old_defn]) is R.PkgInstalled

    result = await iw_manager.install([old_defn], replace=False)
    assert type(result[old_defn]) is R.PkgAlreadyInstalled

    result = await iw_manager.install([new_defn], replace=False)
    assert type(result[new_defn]) is R.PkgNonexistent

    async def resolve(defns, with_deps=False):
        result = await orig_resolve([old_defn])
        pkg = evolve(result[old_defn], slug=new_defn.alias)
        return {new_defn: pkg}

    orig_resolve = iw_manager.resolve
    monkeypatch.setattr(iw_manager, 'resolve', resolve)

    result = await iw_manager.install([new_defn], replace=False)
    assert type(result[new_defn]) is R.PkgAlreadyInstalled


async def test_update_lifecycle_while_varying_retain_defn_strategy(iw_manager: Manager):
    defn = Defn('curse', 'molinari')
    versioned_defn = defn.with_version('100005.97-Release')

    result = (await iw_manager.install([defn], replace=False))[defn]
    assert type(result) is R.PkgInstalled

    result = (await iw_manager.update([defn], retain_defn_strategy=False))[defn]
    assert type(result) is R.PkgUpToDate
    assert result.is_pinned is False

    result = (await iw_manager.update([versioned_defn], retain_defn_strategy=False))[
        versioned_defn
    ]
    assert type(result) is R.PkgUpToDate
    assert result.is_pinned is False

    result = (await iw_manager.update([versioned_defn], retain_defn_strategy=True))[versioned_defn]
    assert type(result) is R.PkgUpdated
    assert result.new_pkg.options.version_eq is True

    result = (await iw_manager.update([defn], retain_defn_strategy=False))[defn]
    assert type(result) is R.PkgUpToDate
    assert result.is_pinned is True

    result = (await iw_manager.update([defn], retain_defn_strategy=True))[defn]
    assert type(result) is R.PkgUpdated
    assert result.new_pkg.options.version_eq is False


async def test_update_reinstalls_corrupted_pkgs(iw_manager: Manager):
    defn = Defn('curse', 'molinari')

    result = (await iw_manager.install([defn], replace=False))[defn]
    assert type(result) is R.PkgInstalled

    folders = [iw_manager.config.addon_dir / f.name for f in result.pkg.folders]

    first_folder = folders[0]
    first_folder.rename(first_folder.with_name('foo'))
    assert not all(f.is_dir() for f in folders)

    result = await iw_manager.update([defn], retain_defn_strategy=False)
    assert type(result[defn]) is R.PkgUpdated
    assert all(f.is_dir() for f in folders)


@pytest.mark.parametrize('keep_folders', [True, False])
async def test_deleting_and_retaining_folders_on_remove(iw_manager: Manager, keep_folders: bool):
    defn = Defn('curse', 'molinari')

    await iw_manager.install([defn], False)
    folders = [iw_manager.config.addon_dir / f.name for f in iw_manager.get_pkg(defn).folders]
    assert all(f.is_dir() for f in folders)

    result = await iw_manager.remove([defn], keep_folders=keep_folders)
    assert type(result[defn]) is R.PkgRemoved
    assert not iw_manager.get_pkg(defn)
    if keep_folders:
        assert all(f.is_dir() for f in folders)
    else:
        assert not any(f.is_dir() for f in folders)


@pytest.mark.parametrize('keep_folders', [True, False])
async def test_removing_pkg_with_missing_folders(iw_manager: Manager, keep_folders: bool):
    defn = Defn('curse', 'molinari')

    result = await iw_manager.install([defn], False)
    folders = [iw_manager.config.addon_dir / f.name for f in result[defn].pkg.folders]
    for folder in folders:
        folder.rename(folder.with_name(f'Not_{folder.name}'))
    assert not any(f.is_dir() for f in folders)

    result = await iw_manager.remove([defn], keep_folders=keep_folders)
    assert type(result[defn]) is R.PkgRemoved
    assert not iw_manager.get_pkg(defn)


async def test_basic_search(iw_manager: Manager):
    limit = 5
    results = await iw_manager.search('molinari', limit=limit)
    assert len(results) <= 5
    assert {('curse', 'molinari'), ('wowi', '13188')} <= {
        (e.source, e.slug or e.id) for e in results
    }


@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
async def test_search_flavour_filtering(iw_manager: Manager):
    results = await iw_manager.search('AtlasLootClassic', limit=5)
    faux_defns = {(e.source, e.slug or e.id) for e in results}
    if iw_manager.config.game_flavour in {
        Flavour.VanillaClassic,
        Flavour.Classic,
    }:
        assert ('curse', 'atlaslootclassic') in faux_defns
    else:
        assert ('curse', 'atlaslootclassic') not in faux_defns


async def test_search_source_filtering(iw_manager: Manager):
    results = await iw_manager.search('molinari', limit=5, sources={'curse'})
    assert all(e.source == 'curse' for e in results)
    assert ('curse', 'molinari') in {(e.source, e.slug) for e in results}


async def test_search_date_filtering(iw_manager: Manager):
    start_date = datetime.datetime.now(tz=datetime.timezone.utc) - datetime.timedelta(days=365)
    results = await iw_manager.search('molinari', limit=5, start_date=start_date)
    assert all(e.last_updated > start_date for e in results)


async def test_search_unknown_source(iw_manager: Manager):
    with pytest.raises(ValueError, match='Unknown source'):
        await iw_manager.search('molinari', limit=5, sources={'foo'})


async def test_search_filter_installed(iw_manager: Manager):
    results = await iw_manager.search('molinari', limit=5, filter_installed='include_only')
    assert not results

    defn = Defn('curse', 'molinari')

    install_result = (await iw_manager.install([defn], False))[defn]
    assert type(install_result) is R.PkgInstalled

    results = await iw_manager.search('molinari', limit=5)
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} <= {
        (e.source, e.slug) for e in results
    }

    results = await iw_manager.search('molinari', limit=5, filter_installed='include_only')
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == {('curse', 'molinari')}

    results = await iw_manager.search('molinari', limit=5, filter_installed='exclude')
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == {('github', 'p3lim-wow/molinari')}

    results = await iw_manager.search(
        'molinari', limit=5, filter_installed='exclude_from_all_sources'
    )
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == set()


async def test_search_prefer_source(iw_manager: Manager):
    results = await iw_manager.search('molinari', limit=5, prefer_source=None)
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} <= {
        (e.source, e.slug) for e in results
    }

    results = await iw_manager.search('molinari', limit=5, prefer_source='github')
    assert {('curse', 'molinari'), ('github', 'p3lim-wow/molinari')} & {
        (e.source, e.slug) for e in results
    } == {('github', 'p3lim-wow/molinari')}


async def test_get_changelog_from_empty_data_url(iw_manager: Manager):
    assert (await iw_manager.get_changelog('github', 'data:,')) == ''


async def test_get_changelog_from_url_encoded_data_url(iw_manager: Manager):
    assert (await iw_manager.get_changelog('github', 'data:,foo%20bar')) == 'foo bar'


async def test_get_malformed_changelog(iw_manager: Manager):
    with pytest.raises(ValueError, match='Unsupported URI with scheme'):
        await iw_manager.get_changelog('github', '')


async def test_get_changelog_from_file_uri(iw_manager: Manager, tmp_path: Path):
    changelog = tmp_path / 'changelog.txt'
    changelog.write_text('test')
    assert (await iw_manager.get_changelog('github', changelog.as_uri())) == 'test'


async def test_get_changelog_from_web_url(iw_manager: Manager):
    assert (
        await iw_manager.get_changelog(
            'curse', 'https://api.curseforge.com/v1/mods/20338/files/3657564/changelog'
        )
    ).startswith('<h3>Changes in')



from __future__ import annotations

from collections.abc import Set
from datetime import datetime
from functools import cached_property
from typing import Any

from attrs import field, frozen
from cattrs import Converter
from cattrs.preconf.json import configure_converter
from typing_extensions import Self

from .config import Flavour
from .utils import bucketise, normalise_names

CATALOGUE_VERSION = 7
COMPUTED_CATALOGUE_VERSION = 4


catalogue_converter = Converter(
    unstruct_collection_overrides={
        Set: sorted,
    }
)
configure_converter(catalogue_converter)

_normalise_name = normalise_names('')


@frozen(kw_only=True)
class AddonKey:
    source: str
    id: str


@frozen(kw_only=True)
class CatalogueEntry(AddonKey):
    slug: str = ''
    name: str
    url: str
    game_flavours: frozenset[Flavour]
    download_count: int
    last_updated: datetime
    folders: list[frozenset[str]] = field(factory=list)
    same_as: list[AddonKey] = field(factory=list)


@frozen(kw_only=True)
class ComputedCatalogueEntry(CatalogueEntry):
    normalised_name: str
    derived_download_score: float


@frozen(kw_only=True)
class Catalogue:
    version: int = CATALOGUE_VERSION
    entries: list[CatalogueEntry]

    @classmethod
    async def collate(cls, start_date: datetime | None) -> Self:
        from .http import init_web_client
        from .manager import Manager

        async with init_web_client(None) as web_client:
            entries = [e for r in Manager.RESOLVERS async for e in r.catalogue(web_client)]
            if start_date is not None:
                entries = [e for e in entries if e.last_updated >= start_date]
            return cls(entries=entries)


@frozen(kw_only=True, slots=False)
class ComputedCatalogue:
    version: int = COMPUTED_CATALOGUE_VERSION
    entries: list[ComputedCatalogueEntry]
    curse_slugs: dict[str, str]

    @classmethod
    def from_base_catalogue(cls, unstructured_base_catalogue: dict[str, Any]) -> Self:
        from ._sources.cfcore import CfCoreResolver
        from ._sources.github import GithubResolver

        normalise_name = _normalise_name

        base_entries = unstructured_base_catalogue['entries']

        most_downloads_per_source = {
            s: max(e['download_count'] for e in i)
            for s, i in bucketise(base_entries, key=lambda e: e['source']).items()
        }
        same_as_from_github = {
            (s['source'], s['id']): [
                e,
                *(i for i in e['same_as'] if i['source'] != s['source']),
            ]
            for e in base_entries
            if e['source'] == GithubResolver.metadata.id and e['same_as']
            for s in e['same_as']
        }

        return catalogue_converter.structure(
            {
                'entries': (
                    e
                    | {
                        'same_as': e['same_as']
                        if e['source'] == GithubResolver.metadata.id
                        else (same_as_from_github.get((e['source'], e['id'])) or e['same_as']),
                        'normalised_name': normalise_name(e['name']),
                        'derived_download_score': 0
                        if e['source'] == GithubResolver.metadata.id
                        else e['download_count'] / most_downloads_per_source[e['source']],
                    }
                    for e in base_entries
                ),
                'curse_slugs': {
                    e['slug']: e['id']
                    for e in base_entries
                    if e['source'] == CfCoreResolver.metadata.id
                },
            },
            cls,
        )

    @cached_property
    def keyed_entries(self) -> dict[tuple[str, str], ComputedCatalogueEntry]:
        return {(e.source, e.id): e for e in self.entries}



# pyright: strict

from __future__ import annotations

import json
import re
from collections.abc import Callable
from functools import cache
from io import BytesIO
from pathlib import Path
from typing import Any
from zipfile import ZipFile

from aiohttp.web import Response
from attr import frozen
from yarl import URL

from instawow import __version__

_HERE = Path(__file__).parent

_match_any = re.compile(r'.*')


def _load_fixture(filename: str):
    return (_HERE / filename).read_bytes()


def _load_json_fixture(filename: str):
    return json.loads(_load_fixture(filename))


@cache
def _make_addon_zip(*folders: str):
    buffer = BytesIO()
    with ZipFile(buffer, 'w') as file:
        for folder in folders:
            file.writestr(f'{folder}/{folder}.toc', b'')

    return buffer.getvalue()


@frozen
class Route:
    url: URL
    response: Callable[[], Response] | dict[str, Any] | str
    path_pattern: re.Pattern[str] | None = None
    method: str = 'GET'
    body_pattern: re.Pattern[str] | None = None
    match_querystring: bool = False
    repeat: float = float('inf')
    case_insensitive: bool = False

    def _make_path_pattern(self):
        if self.path_pattern is not None:
            return self.path_pattern

        if self.case_insensitive:
            return re.compile(fr'^{re.escape(self.url.path_qs)}$', re.IGNORECASE)

        return self.url.path_qs

    def to_aresponses_add_args(self) -> dict[str, Any]:
        return {
            'host_pattern': self.url.host,
            'path_pattern': self._make_path_pattern(),
            'method_pattern': self.method,
            'body_pattern': _match_any if self.body_pattern is None else self.body_pattern,
            'match_querystring': self.match_querystring,
            'repeat': self.repeat,
            'response': self.response() if callable(self.response) else self.response,
        }


def _make_route_dict_entry(route: Route):
    return (route.url, route)


ROUTES = dict(
    map(
        _make_route_dict_entry,
        [
            Route(
                URL('//pypi.org/pypi/instawow/json'),
                {'info': {'version': __version__}},
            ),
            Route(
                URL(
                    '//raw.githubusercontent.com/layday/instawow-data/data/base-catalogue-v7.compact.json'
                ),
                _load_json_fixture('base-catalogue-v7.compact.json'),
            ),
            Route(
                URL('//api.curseforge.com/v1/mods'),
                _load_json_fixture('curse-addon--all.json'),
                method='POST',
            ),
            Route(
                URL('//api.curseforge.com/v1/mods/20338/files'),
                _load_json_fixture('curse-addon-files.json'),
            ),
            Route(
                URL('//api.curseforge.com/v1/mods/20338/files/{id}/changelog'),
                _load_json_fixture('curse-addon-changelog.json'),
                path_pattern=re.compile(r'^/v1/mods/20338/files/(\d+)/changelog$'),
            ),
            Route(
                URL('//edge.forgecdn.net'),
                lambda: Response(body=_make_addon_zip('Molinari')),
                path_pattern=_match_any,
            ),
            Route(
                URL('//api.mmoui.com/v3/game/WOW/filelist.json'),
                _load_json_fixture('wowi-filelist.json'),
            ),
            Route(
                URL('//api.mmoui.com/v3/game/WOW/filedetails/{id}.json'),
                _load_json_fixture('wowi-filedetails.json'),
                path_pattern=re.compile(r'^/v3/game/WOW/filedetails/(\d*)\.json$'),
            ),
            Route(
                URL('//cdn.wowinterface.com'),
                lambda: Response(body=_make_addon_zip('Molinari')),
                path_pattern=_match_any,
            ),
            Route(
                URL('//api.tukui.org/v1/addon/tukui'),
                _load_json_fixture('tukui-ui--tukui.json'),
            ),
            Route(
                URL('//api.tukui.org/v1/addon/elvui'),
                _load_json_fixture('tukui-ui--elvui.json'),
            ),
            Route(
                URL('//api.tukui.org/v1/download/'),
                lambda: Response(body=_make_addon_zip('Tukui')),
                path_pattern=re.compile(r'^/v1/download/'),
            ),
            Route(
                URL('//api.github.com/repos/nebularg/PackagerTest'),
                _load_json_fixture('github-repo-release-json.json'),
            ),
            Route(
                URL('//api.github.com/repos/nebularg/PackagerTest/releases?per_page=10'),
                _load_json_fixture('github-release-release-json.json'),
                match_querystring=True,
            ),
            Route(
                URL('//api.github.com/repos/nebularg/PackagerTest/releases/assets/37156458'),
                _load_json_fixture('github-release-release-json-release-json.json'),
            ),
            Route(
                URL('//api.github.com/repositories/388670'),
                _load_json_fixture('github-repo-molinari.json'),
            ),
            Route(
                URL('//api.github.com/repos/p3lim-wow/Molinari'),
                _load_json_fixture('github-repo-molinari.json'),
                case_insensitive=True,
            ),
            Route(
                URL('//api.github.com/repos/p3lim-wow/Molinari/releases?per_page=10'),
                _load_json_fixture('github-release-molinari.json'),
                case_insensitive=True,
                match_querystring=True,
            ),
            Route(
                URL('//api.github.com/repos/p3lim-wow/Molinari/releases/assets/108929561'),
                _load_json_fixture('github-release-molinari-release-json.json'),
                case_insensitive=True,
            ),
            Route(
                URL('//api.github.com/repos/AdiAddons/AdiBags'),
                _load_json_fixture('github-repo-no-releases.json'),
            ),
            Route(
                URL('//api.github.com/repos/AdiAddons/AdiBags/releases?per_page=10'),
                lambda: Response(body=b'', status=404),
                match_querystring=True,
            ),
            Route(
                URL('//api.github.com/repos/AdiAddons/AdiButtonAuras/releases/tags/2.0.19'),
                _load_json_fixture('github-release-no-assets.json'),
            ),
            Route(
                URL('//api.github.com/repos/layday/foobar'),
                lambda: Response(body=b'', status=404),
            ),
            Route(
                URL('//api.github.com/repos/{x}/{y}/releases/asssets/{z}'),
                lambda: Response(body=_make_addon_zip('Foo')),
                path_pattern=re.compile(r'^/repos(/[^/]*){2}/releases/assets/'),
            ),
            Route(
                URL('//github.com/login/device/code'),
                _load_json_fixture('github-oauth-login-device-code.json'),
                method='POST',
            ),
            Route(
                URL('//github.com/login/oauth/access_token'),
                _load_json_fixture('github-oauth-login-access-token.json'),
                method='POST',
            ),
            Route(
                URL('//api.github.com/repos/28/NoteworthyII'),
                _load_json_fixture('github-repo-no-release-json.json'),
            ),
            Route(
                URL('//api.github.com/repos/28/NoteworthyII/releases?per_page=10'),
                _load_json_fixture('github-release-no-release-json.json'),
                match_querystring=True,
            ),
            Route(
                URL('//addons.wago.io/api/external/addons/_match'),
                _load_json_fixture('wago-match-addons.json'),
                method='POST',
            ),
        ],
    )
)



from __future__ import annotations

import logging
import re

import pytest
from aresponses import ResponsesMockServer
from attrs import evolve
from typing_extensions import assert_never
from yarl import URL

from instawow import results as R
from instawow._sources.cfcore import CfCoreResolver
from instawow._sources.github import GithubResolver
from instawow._sources.wowi import WowiResolver
from instawow.common import Defn, Flavour, Strategy, StrategyValues
from instawow.manager import Manager
from instawow.pkg_models import Pkg
from instawow.resolvers import Resolver


@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
async def test_curse_simple_strategies(iw_manager: Manager):
    flavourful = Defn('curse', 'classiccastbars')
    classics_only = Defn('curse', 'atlaslootclassic')

    results = await iw_manager.resolve([flavourful, classics_only])

    assert type(results[flavourful]) is Pkg

    if iw_manager.config.game_flavour in {Flavour.VanillaClassic, Flavour.Classic}:
        assert type(results[classics_only]) is Pkg
    elif iw_manager.config.game_flavour is Flavour.Retail:
        assert type(results[classics_only]) is R.PkgFilesNotMatching
        assert (
            results[classics_only].message
            == 'no files found for: any_flavour=None; any_release_type=None; version_eq=None'
        )
    else:
        assert_never(iw_manager.config.game_flavour)


async def test_curse_any_flavour_strategy(iw_manager: Manager):
    flavourful = Defn('curse', 'classiccastbars', strategies=StrategyValues(any_flavour=True))
    classics_only = Defn('curse', 'atlaslootclassic', strategies=StrategyValues(any_flavour=True))

    results = await iw_manager.resolve([flavourful, classics_only])
    assert all(type(r) is Pkg for r in results.values())


async def test_curse_version_pinning(iw_manager: Manager):
    defn = Defn('curse', 'molinari', strategies=StrategyValues(version_eq='100005.97-Release'))
    results = await iw_manager.resolve([defn])
    assert results[defn].options.version_eq is True
    assert results[defn].version == '100005.97-Release'


async def test_curse_deps_retrieved(iw_manager: Manager):
    defn = Defn('curse', 'bigwigs-voice-korean')

    results = await iw_manager.resolve([defn], with_deps=True)
    assert {'bigwigs-voice-korean', 'big-wigs'} == {d.slug for d in results.values()}


async def test_curse_changelog_is_url(iw_manager: Manager):
    classiccastbars = Defn('curse', 'classiccastbars')

    results = await iw_manager.resolve([classiccastbars])
    assert re.match(
        r'https://api\.curseforge\.com/v1/mods/\d+/files/\d+/changelog',
        results[classiccastbars].changelog_url,
    )


async def test_wowi_basic(iw_manager: Manager):
    defn = Defn('wowi', '13188-molinari')
    results = await iw_manager.resolve([defn])
    assert type(results[defn]) is Pkg


async def test_wowi_changelog_is_data_url(iw_manager: Manager):
    molinari = Defn('wowi', '13188-molinari')
    results = await iw_manager.resolve([molinari])
    assert results[molinari].changelog_url.startswith('data:,')


@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
async def test_tukui_basic(iw_manager: Manager):
    tukui_suite = Defn('tukui', 'tukui')
    elvui_suite = Defn('tukui', 'elvui')

    results = await iw_manager.resolve([tukui_suite, elvui_suite])

    assert type(results[tukui_suite]) is Pkg
    assert results[tukui_suite].name == 'Tukui'
    assert type(results[elvui_suite]) is Pkg
    assert results[elvui_suite].name == 'ElvUI'


async def test_tukui_changelog_url(iw_manager: Manager):
    ui_suite = Defn('tukui', 'tukui')

    results = await iw_manager.resolve([ui_suite])

    assert results[ui_suite].changelog_url == 'https://api.tukui.org/v1/changelog/tukui#20.37'


async def test_github_basic(iw_manager: Manager):
    release_json = Defn('github', 'nebularg/PackagerTest')
    releaseless = Defn('github', 'AdiAddons/AdiBags')
    nonexistent = Defn('github', 'layday/foobar')

    results = await iw_manager.resolve([release_json, releaseless, nonexistent])

    assert type(results[release_json]) is Pkg
    assert type(results[releaseless]) is R.PkgFilesMissing
    assert results[releaseless].message == 'release not found'
    assert type(results[nonexistent]) is R.PkgNonexistent


async def test_github_changelog_is_data_url(iw_manager: Manager):
    defn = Defn('github', 'p3lim-wow/Molinari')
    results = await iw_manager.resolve([defn])
    assert results[defn].changelog_url.startswith('data:,')


@pytest.mark.parametrize(
    ('iw_config_values', 'flavor', 'interface'),
    [
        (Flavour.Retail, 'mainline', 30400),
        (Flavour.Classic, 'wrath', 90207),
        (Flavour.VanillaClassic, 'classic', 90207),
    ],
    indirect=('iw_config_values',),
)
@pytest.mark.parametrize(
    '_iw_mock_aiohttp_requests',
    [
        {
            URL('//api.github.com/repos/nebularg/PackagerTest'),
            URL('//api.github.com/repos/nebularg/PackagerTest/releases?per_page=10'),
        }
    ],
    indirect=True,
)
async def test_github_flavor_and_interface_mismatch(
    caplog: pytest.LogCaptureFixture,
    aresponses: ResponsesMockServer,
    iw_manager: Manager,
    flavor: str,
    interface: int,
):
    aresponses.add(
        'api.github.com',
        re.compile(r'^/repos/nebularg/PackagerTest/releases/assets/'),
        'GET',
        {
            'releases': [
                {
                    'filename': 'TestGit-v1.9.7.zip',
                    'nolib': False,
                    'metadata': [{'flavor': flavor, 'interface': interface}],
                }
            ]
        },
    )

    defn = Defn('github', 'nebularg/PackagerTest')
    results = await iw_manager.resolve([defn])
    mismatch_result = results[defn]

    assert type(mismatch_result) is R.PkgFilesNotMatching

    (log_record,) = caplog.record_tuples
    assert log_record == (
        'instawow._sources.github',
        logging.INFO,
        f'interface number "{interface}" and flavor "{flavor}" mismatch',
    )


@pytest.mark.parametrize('resolver', Manager.RESOLVERS)
async def test_unsupported_strategies(iw_manager: Manager, resolver: Resolver):
    if resolver.metadata.id not in iw_manager.resolvers:
        pytest.skip('resolver not loaded')

    defn = Defn(resolver.metadata.id, 'foo')
    for strategy in {
        Strategy.AnyFlavour,
        Strategy.AnyReleaseType,
    } - resolver.metadata.strategies:
        strategy_defn = evolve(defn, strategies=StrategyValues(**{strategy: True}))

        results = await iw_manager.resolve([strategy_defn])

        assert type(results[strategy_defn]) is R.PkgStrategiesUnsupported
        assert results[strategy_defn].message == f'strategies are not valid for source: {strategy}'


@pytest.mark.parametrize(
    ('resolver', 'url', 'extracted_alias'),
    [
        (CfCoreResolver, 'https://www.curseforge.com/wow/addons/molinari', 'molinari'),
        (CfCoreResolver, 'https://www.curseforge.com/wow/addons/molinari/download', 'molinari'),
        (WowiResolver, 'https://www.wowinterface.com/downloads/landing.php?fileid=13188', '13188'),
        (WowiResolver, 'https://wowinterface.com/downloads/landing.php?fileid=13188', '13188'),
        (WowiResolver, 'https://www.wowinterface.com/downloads/fileinfo.php?id=13188', '13188'),
        (WowiResolver, 'https://wowinterface.com/downloads/fileinfo.php?id=13188', '13188'),
        (WowiResolver, 'https://www.wowinterface.com/downloads/download13188-Molinari', '13188'),
        (WowiResolver, 'https://wowinterface.com/downloads/download13188-Molinari', '13188'),
        (WowiResolver, 'https://www.wowinterface.com/downloads/info13188-Molinari.html', '13188'),
        (WowiResolver, 'https://wowinterface.com/downloads/info13188-Molinari.html', '13188'),
        (WowiResolver, 'https://www.wowinterface.com/downloads/info13188', '13188'),
        (WowiResolver, 'https://wowinterface.com/downloads/info13188', '13188'),
        (
            GithubResolver,
            'https://github.com/AdiAddons/AdiButtonAuras',
            'AdiAddons/AdiButtonAuras',
        ),
        (
            GithubResolver,
            'https://github.com/AdiAddons/AdiButtonAuras/releases',
            'AdiAddons/AdiButtonAuras',
        ),
    ],
)
def test_get_alias_from_url(resolver: Resolver, url: str, extracted_alias: str):
    assert resolver.get_alias_from_url(URL(url)) == extracted_alias



from __future__ import annotations

from datetime import datetime

import sqlalchemy as sa
from attrs import asdict, field, frozen
from cattrs import Converter
from typing_extensions import Self

from . import pkg_db
from .common import Defn, StrategyValues

_db_pkg_converter = Converter()
_db_pkg_converter.register_structure_hook(datetime, lambda d, _: d)


@frozen(kw_only=True)
class PkgOptions:
    any_flavour: bool
    any_release_type: bool
    version_eq: bool

    @classmethod
    def from_strategy_values(cls, strategies: StrategyValues) -> Self:
        return cls(
            **{k: bool(v) for k, v in asdict(strategies).items()},
        )


@frozen(kw_only=True)
class PkgFolder:
    name: str


@frozen(kw_only=True)
class PkgDep:
    id: str


@frozen(kw_only=True)
class PkgLoggedVersion:
    version: str
    install_time: datetime


@frozen(kw_only=True, eq=False)
class Pkg:
    source: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: datetime
    version: str
    changelog_url: str
    options: PkgOptions  # pkg_options
    folders: list[PkgFolder] = field(factory=list)  # pkg_folder
    deps: list[PkgDep] = field(factory=list)  # pkg_dep
    logged_versions: list[PkgLoggedVersion] = field(factory=list)  # pkg_version_log

    @classmethod
    def from_row_mapping(cls, connection: sa.Connection, row_mapping: sa.RowMapping) -> Self:
        source_and_id = {'pkg_source': row_mapping['source'], 'pkg_id': row_mapping['id']}
        return _db_pkg_converter.structure(
            {
                **row_mapping,
                'options': connection.execute(
                    sa.select(pkg_db.pkg_options).filter_by(**source_and_id)
                )
                .mappings()
                .one(),
                'folders': connection.execute(
                    sa.select(pkg_db.pkg_folder.c.name).filter_by(**source_and_id)
                )
                .mappings()
                .all(),
                'deps': connection.execute(
                    sa.select(pkg_db.pkg_dep.c.id).filter_by(**source_and_id)
                )
                .mappings()
                .all(),
                'logged_versions': connection.execute(
                    sa.select(pkg_db.pkg_version_log)
                    .filter_by(**source_and_id)
                    .order_by(pkg_db.pkg_version_log.c.install_time.desc())
                    .limit(10)
                )
                .mappings()
                .all(),
            },
            cls,
        )

    def insert(self, transaction: sa.Connection) -> None:
        values = asdict(self)
        source_and_id = {'pkg_source': values['source'], 'pkg_id': values['id']}

        transaction.execute(sa.insert(pkg_db.pkg), [values])
        transaction.execute(
            sa.insert(pkg_db.pkg_folder), [{**f, **source_and_id} for f in values['folders']]
        )
        transaction.execute(
            sa.insert(pkg_db.pkg_options), [{**values['options'], **source_and_id}]
        )
        if values['deps']:
            transaction.execute(
                sa.insert(pkg_db.pkg_dep), [{**d, **source_and_id} for d in values['deps']]
            )
        transaction.execute(
            sa.insert(pkg_db.pkg_version_log).prefix_with('OR IGNORE'),
            [{'version': values['version'], **source_and_id}],
        )

    def delete(self, transaction: sa.Connection) -> None:
        transaction.execute(sa.delete(pkg_db.pkg).filter_by(source=self.source, id=self.id))

    def to_defn(self) -> Defn:
        return Defn(
            source=self.source,
            alias=self.slug,
            id=self.id,
            strategies=StrategyValues(
                any_flavour=self.options.any_flavour or None,
                any_release_type=self.options.any_release_type or None,
                version_eq=self.version if self.options.version_eq else None,
            ),
        )



from __future__ import annotations

import re
from io import BytesIO
from zipfile import ZipFile

import aiohttp.hdrs
import aiohttp.web
import pytest
from aresponses import ResponsesMockServer
from yarl import URL

from instawow._sources.github import GithubResolver
from instawow.common import Defn, Flavour
from instawow.manager import Manager
from instawow.results import PkgFilesNotMatching

defn = Defn('github', '28/NoteworthyII')
addon_name = URL(defn.alias).name


ZIPS = {
    'flavoured-toc-only': {
        'toc_files': {
            '_Wrath': b'',
        },
        'flavours': {Flavour.Classic},
    },
    'flavoured-and-unflavoured-toc-without-interface-version': {
        'toc_files': {
            '_Wrath': b'',
            '': b'',
        },
        'flavours': {Flavour.Classic},
    },
    'flavoured-and-unflavoured-toc-with-interface-version': {
        'toc_files': {
            '_Wrath': b'',
            '': b'## Interface: 11300\n',
        },
        'flavours': {Flavour.VanillaClassic, Flavour.Classic},
    },
    'unflavoured-toc-only-without-interface-version': {
        'toc_files': {
            '': b'',
        },
        'flavours': set(),
    },
    'unflavoured-toc-only-with-interface-version': {
        'toc_files': {
            '': b'## Interface: 11300\n',
        },
        'flavours': {Flavour.VanillaClassic},
    },
}


@pytest.fixture(params=ZIPS.values(), ids=list(ZIPS))
def package_json_less_addon(
    request: pytest.FixtureRequest,
):
    addon = BytesIO()

    with ZipFile(addon, 'w') as file:
        for flavour_suffix, content in request.param['toc_files'].items():
            file.writestr(f'{addon_name}/{addon_name}{flavour_suffix}.toc', content)

    return (addon.getvalue(), request.param['flavours'])


@pytest.mark.parametrize(
    '_iw_mock_aiohttp_requests',
    [
        {
            URL(f'//api.github.com/repos/{defn.alias}'),
            URL(f'//api.github.com/repos/{defn.alias}/releases?per_page=10'),
        }
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    'iw_config_values',
    Flavour,
    indirect=True,
)
async def test_package_json_less_addon(
    aresponses: ResponsesMockServer,
    iw_manager: Manager,
    package_json_less_addon: tuple[bytes, set[Flavour]],
):
    async def handle_request(request: aiohttp.web.Request):
        if aiohttp.hdrs.RANGE in request.headers:
            raise aiohttp.web.HTTPRequestRangeNotSatisfiable

        response = aiohttp.web.Response(body=addon)
        await response.prepare(request)
        return response

    aresponses.add(
        'api.github.com',
        re.compile(r'^/repos(/[^/]*){2}/releases/assets/'),
        'get',
        handle_request,
        repeat=aresponses.INFINITY,
    )

    addon, flavours = package_json_less_addon
    try:
        await GithubResolver(iw_manager).resolve_one(defn, None)
    except PkgFilesNotMatching:
        assert iw_manager.config.game_flavour not in flavours
    else:
        assert iw_manager.config.game_flavour in flavours



from __future__ import annotations

import asyncio
import json
import os
from contextlib import suppress
from functools import partial
from typing import Any

import anyio.from_thread
import anyio.to_thread
import toga
import toga.constants
import toga.style
from loguru import logger
from typing_extensions import Self

from . import json_rpc_server


class InstawowApp(toga.App):
    def __init__(self, debug: bool, **kwargs: object) -> None:
        super().__init__(
            formal_name='instawow-gui',
            app_id='org.instawow.instawow_gui',
            app_name='instawow_gui',
            icon='resources/instawow_gui',
            **kwargs,
        )
        self._debug = debug

    def startup(self) -> None:
        self.main_window = toga.MainWindow(title=self.formal_name, size=(800, 600))

        if os.name == 'nt':
            import ctypes

            # Enable high DPI support.
            ctypes.windll.user32.SetProcessDPIAware()

        self.main_window.content = web_view = toga.WebView(style=toga.style.Pack(flex=1))

        if self._debug:
            with suppress(AttributeError):
                web_view._impl.native.configuration.preferences.setValue(
                    True, forKey='developerExtrasEnabled'
                )

        if os.name == 'nt':
            from toga_winforms.widgets.webview import TogaWebBrowser

            def configure_webview2(sender: TogaWebBrowser, event_args: Any):
                if event_args.IsSuccess:
                    sender.CoreWebView2.Settings.AreBrowserAcceleratorKeysEnabled = False

            web_view._impl.native.CoreWebView2InitializationCompleted += configure_webview2

        def dispatch_js_keyboard_event(_: toga.Command, *, action: str):
            event_args = json.dumps(
                {'detail': {'action': action}, 'bubbles': True, 'cancelable': True}
            )
            web_view.invoke_javascript(
                f'document.dispatchEvent(new CustomEvent("togaSimulateKeypress", {event_args}));'
            )

        self.commands.add(
            toga.Command(
                partial(dispatch_js_keyboard_event, action='toggleSearchFilter'),
                text='Toggle Search Filter',
                shortcut=toga.Key.MOD_1 + toga.Key.G,
                group=toga.Group.EDIT,
                section=20,
                order=10,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewInstalled'),
                text='Installed',
                shortcut=toga.Key.MOD_1 + toga.Key.L,
                group=toga.Group.WINDOW,
                section=20,
                order=10,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewReconcile'),
                text='Unreconciled',
                group=toga.Group.WINDOW,
                section=20,
                order=20,
            ),
            toga.Command(
                partial(dispatch_js_keyboard_event, action='activateViewSearch'),
                text='Search',
                shortcut=toga.Key.MOD_1 + toga.Key.F,
                group=toga.Group.WINDOW,
                section=20,
                order=30,
            ),
        )

        async def startup(app: Self):
            async with anyio.from_thread.BlockingPortal() as portal:

                def run_json_rpc_server():
                    async def run():
                        web_app = await json_rpc_server.create_app((app.main_window, portal))
                        server_url, serve_forever = await json_rpc_server.run_app(web_app)

                        logger.debug(f'JSON-RPC server running on {server_url}')

                        set_server_url = partial(setattr, web_view, 'url')
                        portal.call(set_server_url, str(server_url))

                        await serve_forever()

                    # We don't want to inherit the parent thread's event loop policy,
                    # i.e. the rubicon loop on macOS.
                    policy = asyncio.DefaultEventLoopPolicy()
                    policy.new_event_loop().run_until_complete(run())

                await anyio.to_thread.run_sync(run_json_rpc_server)

                await portal.sleep_until_stopped()

        self.add_background_task(startup)
        self.main_window.show()



from __future__ import annotations

import contextvars as cv
import enum
import json
from collections.abc import (
    Awaitable,
    Callable,
    Collection,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    Set,
)
from contextlib import AbstractAsyncContextManager, asynccontextmanager, contextmanager
from datetime import datetime, timedelta
from functools import cached_property, lru_cache, wraps
from itertools import chain, filterfalse, product, repeat, starmap
from pathlib import Path, PurePath
from shutil import move
from tempfile import NamedTemporaryFile
from typing import Literal, TypeVar

import sqlalchemy as sa
from attrs import evolve
from loguru import logger
from typing_extensions import Concatenate, Never, ParamSpec, Self, TypeAlias
from yarl import URL

from . import http, pkg_db, pkg_models
from . import results as R
from ._sources.cfcore import CfCoreResolver
from ._sources.github import GithubResolver
from ._sources.instawow import InstawowResolver
from ._sources.tukui import TukuiResolver
from ._sources.wago import WagoResolver
from ._sources.wowi import WowiResolver
from .cataloguer import (
    CATALOGUE_VERSION,
    ComputedCatalogue,
    ComputedCatalogueEntry,
)
from .common import Defn, Strategy
from .config import Config
from .http import make_generic_progress_ctx, make_pkg_progress_ctx
from .plugins import load_plugins
from .resolvers import HeadersIntent, Resolver
from .utils import (
    StrEnum,
    WeakValueDefaultDictionary,
    bucketise,
    chain_dict,
    file_uri_to_path,
    find_addon_zip_tocs,
    gather,
    is_file_uri,
    make_zip_member_filter_fn,
    normalise_names,
    shasum,
    time_op,
    trash,
    uniq,
)
from .utils import run_in_thread as t

_P = ParamSpec('_P')
_T = TypeVar('_T')

_AsyncNamedTemporaryFile = t(NamedTemporaryFile)
_move_async = t(move)


def bucketise_results(
    values: Iterable[tuple[Defn, R.AnyResult[_T]]],
) -> tuple[Mapping[Defn, _T], Mapping[Defn, R.AnyResult[Never]]]:
    ts: dict[Defn, _T] = {}
    errors: dict[Defn, R.AnyResult[Never]] = {}

    for defn, value in values:
        if isinstance(value, (R.ManagerError, R.InternalError)):
            errors[defn] = value
        else:
            ts[defn] = value

    return ts, errors


@asynccontextmanager
async def _open_temp_writer_async():
    fh = await _AsyncNamedTemporaryFile(delete=False)
    path = Path(fh.name)
    try:
        yield (path, t(fh.write))
    except BaseException:
        await t(fh.close)()
        await t(path.unlink)()
        raise
    else:
        await t(fh.close)()


@contextmanager
def _open_pkg_archive(path: PurePath):
    from zipfile import ZipFile

    with ZipFile(path) as archive:

        def extract(parent: Path) -> None:
            archive.extractall(parent, members=filter(make_zip_member_filter_fn(base_dirs), names))

        names = archive.namelist()
        base_dirs = {h for _, h in find_addon_zip_tocs(names)}
        yield (base_dirs, extract)


@object.__new__
class _DummyResolver:
    async def resolve(self, defns: Sequence[Defn]) -> dict[Defn, R.AnyResult[pkg_models.Pkg]]:
        return dict.fromkeys(defns, R.PkgSourceInvalid())

    async def get_changelog(self, uri: URL) -> str:
        raise R.PkgSourceInvalid


async def capture_manager_exc_async(
    awaitable: Awaitable[_T],
) -> R.AnyResult[_T]:
    "Capture and log an exception raised in a coroutine."
    from aiohttp import ClientError

    try:
        return await awaitable
    except (R.ManagerError, R.InternalError) as error:
        return error
    except ClientError as error:
        logger.opt(exception=True).info('network error')
        return R.InternalError(error)
    except BaseException as error:
        logger.exception('unclassed error')
        return R.InternalError(error)


class _DummyLock:
    async def __aenter__(self):
        pass

    async def __aexit__(self, *args: object):
        pass


def _with_lock(
    lock_name: str,
    *,
    manager_bound: bool = True,
):
    def outer(
        coro_fn: Callable[Concatenate[Manager, _P], Awaitable[_T]]
    ) -> Callable[Concatenate[Manager, _P], Awaitable[_T]]:
        @wraps(coro_fn)
        async def inner(self: Manager, *args: _P.args, **kwargs: _P.kwargs):
            async with self.locks[(lock_name, id(self)) if manager_bound else lock_name]:
                return await coro_fn(self, *args, **kwargs)

        return inner

    return outer


class _StandardLocks(StrEnum):
    LoadCatalogue = enum.auto()
    MutatePkgs = enum.auto()
    DownloadPkg = enum.auto()


class _Resolvers(dict[str, Resolver]):
    @cached_property
    def priority_dict(self) -> _ResolverPriorityDict:
        return _ResolverPriorityDict(self)

    @cached_property
    def addon_toc_key_and_id_pairs(self) -> Collection[tuple[str, str]]:
        return [
            (r.metadata.addon_toc_key, r.metadata.id)
            for r in self.values()
            if r.metadata.addon_toc_key
        ]


class _ResolverPriorityDict(dict[str, float]):
    def __init__(self, resolvers: _Resolvers) -> None:
        super().__init__((n, i) for i, n in enumerate(resolvers))

    def __missing__(self, key: str) -> float:
        return float('inf')


_web_client = cv.ContextVar[http.ClientSessionType]('_web_client')

LocksType: TypeAlias = Mapping[object, AbstractAsyncContextManager[None]]

_locks = cv.ContextVar[LocksType]('_locks', default=WeakValueDefaultDictionary(_DummyLock))


def contextualise(
    *,
    web_client: http.ClientSessionType | None = None,
    locks: LocksType | None = None,
) -> None:
    "Set variables for the current context."
    if web_client is not None:
        _web_client.set(web_client)
    if locks is not None:
        _locks.set(locks)


@lru_cache(1)
def _parse_catalogue(raw_catalogue: bytes):
    with time_op(lambda t: logger.debug(f'parsed catalogue in {t:.3f}s')):
        return ComputedCatalogue.from_base_catalogue(json.loads(raw_catalogue))


class Manager:
    RESOLVERS: Sequence[type[Resolver]] = [
        GithubResolver,
        CfCoreResolver,
        WowiResolver,
        TukuiResolver,
        InstawowResolver,
        WagoResolver,
    ]
    'Default resolvers.'

    _base_catalogue_url = (
        f'https://raw.githubusercontent.com/layday/instawow-data/data/'
        f'base-catalogue-v{CATALOGUE_VERSION}.compact.json'
    )
    _catalogue_ttl = timedelta(hours=4)

    _normalise_search_terms = staticmethod(normalise_names(''))

    def __init__(
        self,
        config: Config,
        database: sa.Engine,
    ) -> None:
        self.config: Config = config
        self.database: sa.Engine = database

        builtin_resolver_classes = list(self.RESOLVERS)

        for resolver, access_token in (
            (r, getattr(self.config.global_config.access_tokens, r.requires_access_token, None))
            for r in self.RESOLVERS
            if r.requires_access_token is not None
        ):
            if access_token is None:
                builtin_resolver_classes.remove(resolver)

        plugin_hook = load_plugins()
        resolver_classes = chain(
            (r for g in plugin_hook.instawow_add_resolvers() for r in g), builtin_resolver_classes
        )
        self.resolvers = _Resolvers((r.metadata.id, r(self)) for r in resolver_classes)

    @classmethod
    def from_config(cls, config: Config) -> Self:
        "Instantiate the manager from a configuration object."
        return cls(config, pkg_db.prepare_database(config.db_uri))

    @property
    def locks(self) -> LocksType:
        "Lock factory used to synchronise async operations."
        return _locks.get()

    @property
    def web_client(self) -> http.ClientSessionType:
        return _web_client.get()

    def pair_uri(self, value: str) -> tuple[str, str] | None:
        "Attempt to extract a valid ``Defn`` source and alias from a URL."
        aliases_from_url = (
            (r.metadata.id, a)
            for r in self.resolvers.values()
            for a in (r.get_alias_from_url(URL(value)),)
            if a
        )
        return next(aliases_from_url, None)

    def check_pkg_exists(self, defn: Defn) -> bool:
        "Check that a package exists in the database."
        with self.database.connect() as connection:
            return (
                connection.execute(
                    sa.select(sa.text('1'))
                    .select_from(pkg_db.pkg)
                    .where(
                        pkg_db.pkg.c.source == defn.source,
                        (pkg_db.pkg.c.id == defn.alias)
                        | (pkg_db.pkg.c.id == defn.id)
                        | (sa.func.lower(pkg_db.pkg.c.slug) == sa.func.lower(defn.alias)),
                    )
                ).scalar()
                == 1
            )

    def get_pkg(self, defn: Defn, partial_match: bool = False) -> pkg_models.Pkg | None:
        "Retrieve a package from the database."
        with self.database.connect() as connection:
            maybe_row_mapping = (
                connection.execute(
                    sa.select(pkg_db.pkg).where(
                        pkg_db.pkg.c.source == defn.source,
                        (pkg_db.pkg.c.id == defn.alias)
                        | (pkg_db.pkg.c.id == defn.id)
                        | (sa.func.lower(pkg_db.pkg.c.slug) == sa.func.lower(defn.alias)),
                    )
                )
                .mappings()
                .one_or_none()
            )
            if maybe_row_mapping is None and partial_match:
                maybe_row_mapping = (
                    connection.execute(
                        sa.select(pkg_db.pkg)
                        .where(pkg_db.pkg.c.slug.contains(defn.alias))
                        .order_by(pkg_db.pkg.c.name)
                    )
                    .mappings()
                    .first()
                )
            if maybe_row_mapping is not None:
                return pkg_models.Pkg.from_row_mapping(connection, maybe_row_mapping)

    @t
    def _install_pkg(self, pkg: pkg_models.Pkg, archive: Path, replace: bool) -> R.PkgInstalled:
        with _open_pkg_archive(archive) as (top_level_folders, extract):
            with self.database.connect() as connection:
                installed_conflicts = connection.execute(
                    sa.select(pkg_db.pkg)
                    .distinct()
                    .join(pkg_db.pkg_folder)
                    .where(pkg_db.pkg_folder.c.name.in_(top_level_folders))
                ).all()
            if installed_conflicts:
                raise R.PkgConflictsWithInstalled(installed_conflicts)

            if replace:
                trash(
                    (self.config.addon_dir / f for f in top_level_folders),
                    dest=self.config.global_config.temp_dir,
                    missing_ok=True,
                )
            else:
                unreconciled_conflicts = top_level_folders & {
                    f.name for f in self.config.addon_dir.iterdir()
                }
                if unreconciled_conflicts:
                    raise R.PkgConflictsWithUnreconciled(unreconciled_conflicts)

            extract(self.config.addon_dir)

        pkg = evolve(
            pkg, folders=[pkg_models.PkgFolder(name=f) for f in sorted(top_level_folders)]
        )
        with self.database.begin() as transaction:
            pkg.insert(transaction)

        return R.PkgInstalled(pkg)

    @t
    def _update_pkg(
        self, old_pkg: pkg_models.Pkg, new_pkg: pkg_models.Pkg, archive: Path
    ) -> R.PkgUpdated:
        with _open_pkg_archive(archive) as (top_level_folders, extract):
            with self.database.connect() as connection:
                installed_conflicts = connection.execute(
                    sa.select(pkg_db.pkg)
                    .distinct()
                    .join(pkg_db.pkg_folder)
                    .where(
                        pkg_db.pkg_folder.c.pkg_source != new_pkg.source,
                        pkg_db.pkg_folder.c.pkg_id != new_pkg.id,
                        pkg_db.pkg_folder.c.name.in_(top_level_folders),
                    )
                ).all()
            if installed_conflicts:
                raise R.PkgConflictsWithInstalled(installed_conflicts)

            unreconciled_conflicts = top_level_folders - {f.name for f in old_pkg.folders} & {
                f.name for f in self.config.addon_dir.iterdir()
            }
            if unreconciled_conflicts:
                raise R.PkgConflictsWithUnreconciled(unreconciled_conflicts)

            trash(
                (self.config.addon_dir / f.name for f in old_pkg.folders),
                dest=self.config.global_config.temp_dir,
                missing_ok=True,
            )
            extract(self.config.addon_dir)

        new_pkg = evolve(
            new_pkg, folders=[pkg_models.PkgFolder(name=f) for f in sorted(top_level_folders)]
        )
        with self.database.begin() as transaction:
            old_pkg.delete(transaction)
            new_pkg.insert(transaction)

        return R.PkgUpdated(old_pkg, new_pkg)

    @t
    def _remove_pkg(self, pkg: pkg_models.Pkg, keep_folders: bool) -> R.PkgRemoved:
        if not keep_folders:
            trash(
                (self.config.addon_dir / f.name for f in pkg.folders),
                dest=self.config.global_config.temp_dir,
                missing_ok=True,
            )

        with self.database.begin() as transaction:
            pkg.delete(transaction)

        return R.PkgRemoved(pkg)

    @_with_lock(_StandardLocks.LoadCatalogue, manager_bound=False)
    async def synchronise(self) -> ComputedCatalogue:
        "Fetch the catalogue from the interwebs and load it."
        async with self.web_client.get(
            self._base_catalogue_url,
            expire_after=self._catalogue_ttl,
            raise_for_status=True,
            trace_request_ctx=make_generic_progress_ctx('Synchronising catalogue'),
        ) as response:
            raw_catalogue = await response.read()
        return _parse_catalogue(raw_catalogue)

    async def find_equivalent_pkg_defns(
        self, pkgs: Collection[pkg_models.Pkg]
    ) -> dict[pkg_models.Pkg, list[Defn]]:
        "Given a list of packages, find ``Defn``s of each package from other sources."
        from .matchers import AddonFolder

        catalogue = await self.synchronise()

        @t
        def collect_addon_folders():
            return {
                p: frozenset(
                    a
                    for f in p.folders
                    for a in (
                        AddonFolder.from_addon_path(
                            self.config.game_flavour, self.config.addon_dir / f.name
                        ),
                    )
                    if a
                )
                for p in pkgs
            }

        async def collect_hashed_folder_defns():
            matches = [
                m
                for r in self.resolvers.values()
                for m in await r.get_folder_hash_matches(
                    [a for p, f in folders_per_pkg.items() if p.source != r.metadata.id for a in f]
                )
            ]

            defns_per_pkg: dict[pkg_models.Pkg, frozenset[Defn]] = {
                p: frozenset() for p in folders_per_pkg
            }
            for (pkg, orig_folders), (defn, matched_folders) in product(
                folders_per_pkg.items(), matches
            ):
                if orig_folders == matched_folders:
                    defns_per_pkg[pkg] |= frozenset((defn,))

            return defns_per_pkg

        def get_catalogue_defns(pkg: pkg_models.Pkg) -> frozenset[Defn]:
            entry = catalogue.keyed_entries.get((pkg.source, pkg.id))
            if entry:
                return frozenset(Defn(s.source, s.id) for s in entry.same_as)
            else:
                return frozenset()

        def get_addon_toc_defns(pkg_source: str, addon_folders: Collection[AddonFolder]):
            return frozenset(
                d
                for a in addon_folders
                for d in a.get_defns_from_toc_keys(self.resolvers.addon_toc_key_and_id_pairs)
                if d.source != pkg_source
            )

        folders_per_pkg = await collect_addon_folders()

        with time_op(lambda t: logger.debug(f'hashed folder matches found in {t:.3f}s')):
            hashed_folder_defns = await collect_hashed_folder_defns()

        return {
            p: sorted(d, key=lambda d: self.resolvers.priority_dict[d.source])
            for p in pkgs
            for d in (
                get_catalogue_defns(p)
                | get_addon_toc_defns(p.source, folders_per_pkg[p])
                | hashed_folder_defns[p],
            )
            if d
        }

    async def _resolve_deps(
        self, results: Collection[R.AnyResult[pkg_models.Pkg]]
    ) -> Mapping[Defn, R.AnyResult[pkg_models.Pkg]]:
        """Resolve package dependencies.

        The resolver will not follow dependencies
        more than one level deep.  This is to avoid unnecessary
        complexity for something that I would never expect to
        encounter in the wild.
        """
        pkgs = [r for r in results if isinstance(r, pkg_models.Pkg)]
        dep_defns = uniq(
            filterfalse(
                {(p.source, p.id) for p in pkgs}.__contains__,
                ((p.source, d.id) for p in pkgs for d in p.deps),
            )
        )
        if not dep_defns:
            return {}

        deps = await self.resolve(list(starmap(Defn, dep_defns)))
        pretty_deps = {
            evolve(d, alias=r.slug) if isinstance(r, pkg_models.Pkg) else d: r
            for d, r in deps.items()
        }
        return pretty_deps

    async def resolve(
        self, defns: Collection[Defn], with_deps: bool = False
    ) -> Mapping[Defn, R.AnyResult[pkg_models.Pkg]]:
        "Resolve definitions into packages."
        if not defns:
            return {}

        defns_by_source = bucketise(defns, key=lambda v: v.source)
        results = await gather(
            (self.resolvers.get(s, _DummyResolver).resolve(b) for s, b in defns_by_source.items()),
            capture_manager_exc_async,
        )
        results_by_defn = chain_dict(
            defns,
            R.ManagerError(),
            *(
                r.items() if isinstance(r, dict) else zip(d, repeat(r))
                for d, r in zip(defns_by_source.values(), results)
            ),
        )
        if with_deps:
            results_by_defn.update(await self._resolve_deps(results_by_defn.values()))
        return results_by_defn

    async def get_changelog(self, source: str, uri: str) -> str:
        "Retrieve a changelog from a URI."
        return await self.resolvers.get(source, _DummyResolver).get_changelog(URL(uri))

    async def search(
        self,
        search_terms: str,
        *,
        limit: int,
        sources: Set[str] = frozenset(),
        prefer_source: str | None = None,
        start_date: datetime | None = None,
        filter_installed: Literal[
            'ident', 'include_only', 'exclude', 'exclude_from_all_sources'
        ] = 'ident',
    ) -> list[ComputedCatalogueEntry]:
        "Search the catalogue for packages by name."
        import rapidfuzz

        catalogue = await self.synchronise()

        ew = 0.5
        dw = 1 - ew

        threshold = 0 if search_terms == '*' else 70

        if sources:
            unknown_sources = sources - self.resolvers.keys()
            if unknown_sources:
                raise ValueError(f'Unknown sources: {", ".join(unknown_sources)}')

        if prefer_source and prefer_source not in self.resolvers:
            raise ValueError(f'Unknown preferred source: {prefer_source}')

        def get_installed_pkg_keys():
            with self.database.connect() as connection:
                return (
                    connection.execute(sa.select(pkg_db.pkg.c.source, pkg_db.pkg.c.id))
                    .tuples()
                    .all()
                )

        def make_filter_fns() -> Iterator[Callable[[ComputedCatalogueEntry], bool]]:
            yield lambda e: self.config.game_flavour in e.game_flavours

            if sources:
                yield lambda e: e.source in sources

            if start_date is not None:
                start_date_ = start_date
                yield lambda e: e.last_updated >= start_date_

            if filter_installed in {'exclude', 'exclude_from_all_sources'}:
                installed_pkg_keys = set(get_installed_pkg_keys())
                if filter_installed == 'exclude_from_all_sources':
                    installed_pkg_keys |= {
                        (s.source, s.id)
                        for k in installed_pkg_keys
                        for e in (catalogue.keyed_entries.get(k),)
                        if e
                        for s in e.same_as
                    }

                yield lambda e: (e.source, e.id) not in installed_pkg_keys

        filter_fns = list(make_filter_fns())

        entries = catalogue.entries

        if prefer_source:
            entries = (e for e in entries if not any(s.source == prefer_source for s in e.same_as))

        if filter_installed == 'include_only':
            installed_pkg_keys = get_installed_pkg_keys()
            entries = (
                e for k in installed_pkg_keys for e in (catalogue.keyed_entries.get(k),) if e
            )

        s = self._normalise_search_terms(search_terms)

        tokens_to_entries = bucketise(
            ((e.normalised_name, e) for e in entries if all(f(e) for f in filter_fns)),
            key=lambda v: v[0],
        )
        matches = rapidfuzz.process.extract(
            s,
            list(tokens_to_entries),
            scorer=rapidfuzz.fuzz.WRatio,
            limit=limit * 2,
            score_cutoff=threshold,
        )
        weighted_entries = sorted(
            (
                (-((s / 100) * ew + e.derived_download_score * dw), e)
                for m, s, _ in matches
                for _, e in tokens_to_entries[m]
            ),
            key=lambda v: v[0],
        )
        return [e for _, e in weighted_entries[:limit]]

    async def _download_pkg_archive(self, pkg: pkg_models.Pkg, *, chunk_size: int = 4096):
        if is_file_uri(pkg.download_url):
            return Path(file_uri_to_path(pkg.download_url))

        async with self.locks[_StandardLocks.DownloadPkg, pkg.download_url]:
            dest = self.config.global_config.cache_dir / shasum(pkg.download_url)
            if not await t(dest.exists)():
                async with self.web_client.get(
                    pkg.download_url,
                    headers=await self.resolvers[pkg.source].make_request_headers(
                        intent=HeadersIntent.Download
                    ),
                    raise_for_status=True,
                    trace_request_ctx=make_pkg_progress_ctx(self.config.profile, pkg),
                ) as response, _open_temp_writer_async() as (
                    temp_path,
                    write,
                ):
                    async for chunk in response.content.iter_chunked(chunk_size):
                        await write(chunk)

                await _move_async(temp_path, dest)

            return dest

    @t
    def _check_installed_pkg_integrity(self, pkg: pkg_models.Pkg) -> bool:
        return all((self.config.addon_dir / p.name).exists() for p in pkg.folders)

    async def _should_update_pkg(self, old_pkg: pkg_models.Pkg, new_pkg: pkg_models.Pkg) -> bool:
        return old_pkg.version != new_pkg.version or (
            not await self._check_installed_pkg_integrity(old_pkg)
        )

    @_with_lock(_StandardLocks.MutatePkgs)
    async def install(
        self, defns: Sequence[Defn], replace: bool, dry_run: bool = False
    ) -> Mapping[Defn, R.AnyResult[R.PkgInstalled]]:
        "Install packages from a definition list."

        # We'll weed out installed deps from the results after resolving -
        # doing it this way isn't particularly efficient but avoids having to
        # deal with local state in ``resolve``
        resolve_results = await self.resolve(
            [d for d in defns if not self.check_pkg_exists(d)], with_deps=True
        )
        pkgs, resolve_errors = bucketise_results(resolve_results.items())
        new_pkgs = {d: p for d, p in pkgs.items() if not self.check_pkg_exists(p.to_defn())}

        results = dict.fromkeys(defns, R.PkgAlreadyInstalled()) | resolve_errors

        if dry_run:
            return results | {d: R.PkgInstalled(p, dry_run=True) for d, p in new_pkgs.items()}

        download_results = zip(
            new_pkgs,
            await gather(
                (self._download_pkg_archive(r) for r in new_pkgs.values()),
                capture_manager_exc_async,
            ),
        )
        archive_paths, download_errors = bucketise_results(download_results)

        return (
            results
            | download_errors
            | {
                d: await capture_manager_exc_async(self._install_pkg(new_pkgs[d], a, replace))
                for d, a in archive_paths.items()
            }
        )

    @_with_lock(_StandardLocks.MutatePkgs)
    async def update(
        self, defns: Sequence[Defn], retain_defn_strategy: bool, dry_run: bool = False
    ) -> Mapping[Defn, R.AnyResult[R.PkgInstalled | R.PkgUpdated]]:
        """Update installed packages from a definition list.

        A ``retain_defn_strategy`` value of false will instruct ``update``
        to extract the strategy from the installed package; otherwise
        the ``Defn`` strategy will be used.
        """

        defns_to_pkgs = {d: p for d in defns for p in (self.get_pkg(d),) if p}
        resolve_defns = {
            # Attach the source ID to each ``Defn`` from the
            # corresponding installed package.  Using the ID has the benefit
            # of resolving installed-but-renamed packages - the slug is
            # transient but the ID isn't
            evolve(d, id=p.id) if retain_defn_strategy else p.to_defn(): d
            for d, p in defns_to_pkgs.items()
        }
        resolve_results = await self.resolve(resolve_defns, with_deps=True)
        # Discard the reconstructed ``Defn``s
        orig_defn_resolve_results = (
            (resolve_defns.get(d, d), r) for d, r in resolve_results.items()
        )
        pkgs, resolve_errors = bucketise_results(orig_defn_resolve_results)

        updatables = {
            d: (o, n)
            for d, n in pkgs.items()
            for o in (defns_to_pkgs.get(d),)
            if not o or await self._should_update_pkg(o, n)
        }

        results = (
            dict.fromkeys(defns, R.PkgNotInstalled())
            | resolve_errors
            | {
                d: R.PkgUpToDate(is_pinned=pkgs[d].options.version_eq)
                for d in pkgs.keys() - updatables.keys()
            }
        )

        if dry_run:
            return results | {
                d: R.PkgUpdated(o, n, dry_run=True) if o else R.PkgInstalled(n, dry_run=True)
                for d, (o, n) in updatables.items()
            }

        download_results = zip(
            updatables,
            await gather(
                (self._download_pkg_archive(n) for _, n in updatables.values()),
                capture_manager_exc_async,
            ),
        )
        archive_paths, download_errors = bucketise_results(download_results)

        return (
            results
            | download_errors
            | {
                d: await capture_manager_exc_async(
                    self._update_pkg(o, n, a) if o else self._install_pkg(n, a, False)
                )
                for d, a in archive_paths.items()
                for o, n in (updatables[d],)
            }
        )

    @_with_lock(_StandardLocks.MutatePkgs)
    async def remove(
        self, defns: Sequence[Defn], keep_folders: bool
    ) -> Mapping[Defn, R.AnyResult[R.PkgRemoved]]:
        "Remove packages by their definition."
        return {
            d: (
                await capture_manager_exc_async(self._remove_pkg(p, keep_folders))
                if p
                else R.PkgNotInstalled()
            )
            for d in defns
            for p in (self.get_pkg(d),)
        }

    @_with_lock(_StandardLocks.MutatePkgs)
    async def pin(self, defns: Sequence[Defn]) -> Mapping[Defn, R.AnyResult[R.PkgInstalled]]:
        """Pin and unpin installed packages.

        instawow does not have true pinning.  This flips ``Strategy.VersionEq``
        on for installed packages from sources that support it.
        The net effect is the same as if the package
        had been reinstalled with the ``VersionEq`` strategy.
        """

        @t
        def pin(defn: Defn) -> R.PkgInstalled:
            resolver = self.resolvers.get(defn.source)
            if resolver is None:
                raise R.PkgSourceInvalid

            if Strategy.VersionEq not in resolver.metadata.strategies:
                raise R.PkgStrategiesUnsupported({Strategy.VersionEq})

            pkg = self.get_pkg(defn)
            if not pkg:
                raise R.PkgNotInstalled

            version = defn.strategies.version_eq
            if version and pkg.version != version:
                R.PkgFilesNotMatching(defn.strategies)

            with self.database.begin() as transaction:
                transaction.execute(
                    sa.update(pkg_db.pkg_options)
                    .filter_by(pkg_source=pkg.source, pkg_id=pkg.id)
                    .values(version_eq=version is not None)
                )

            new_pkg = evolve(pkg, options=evolve(pkg.options, version_eq=version is not None))
            return R.PkgInstalled(new_pkg)

        return {d: await capture_manager_exc_async(pin(d)) for d in defns}



from __future__ import annotations

from collections.abc import Collection
from typing import Any, Final, Protocol, TypeVar

from attrs import asdict
from typing_extensions import TypeAlias

from . import pkg_models
from .common import Strategy, StrategyValues

_T = TypeVar('_T')

AnyResult: TypeAlias = '_T | ManagerError | InternalError'


class Result(Protocol):
    @property
    def message(self) -> str:
        ...


class _SuccessResult:
    status: Final = 'success'


class PkgInstalled(Result, _SuccessResult):
    def __init__(self, pkg: pkg_models.Pkg, *, dry_run: bool = False) -> None:
        super().__init__()
        self.pkg = pkg
        self.dry_run = dry_run

    @property
    def message(self) -> str:
        return f'{"would have installed" if self.dry_run else "installed"} {self.pkg.version}'


class PkgUpdated(Result, _SuccessResult):
    def __init__(
        self, old_pkg: pkg_models.Pkg, new_pkg: pkg_models.Pkg, *, dry_run: bool = False
    ) -> None:
        super().__init__()
        self.old_pkg = old_pkg
        self.new_pkg = new_pkg
        self.dry_run = dry_run

    @property
    def message(self) -> str:
        message = (
            f'{"would have updated" if self.dry_run else "updated"}'
            f' {self.old_pkg.version} to {self.new_pkg.version}'
        )

        if self.old_pkg.slug != self.new_pkg.slug:
            message += f' with new slug {self.new_pkg.slug!r}'

        if self.old_pkg.options != self.new_pkg.options:
            old_strategies = asdict(self.old_pkg.to_defn().strategies)
            new_strategies = asdict(self.new_pkg.to_defn().strategies)
            message += ' with new strategies: '
            message += '; '.join(
                f'{s}={v!r}' for s, v in new_strategies.items() - old_strategies.items()
            )

        return message


class PkgRemoved(Result, _SuccessResult):
    def __init__(self, old_pkg: pkg_models.Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg

    @property
    def message(self) -> str:
        return 'removed'


class ManagerError(Result, Exception):
    status: Final = 'failure'

    @property
    def message(self) -> str:
        raise NotImplementedError


class PkgAlreadyInstalled(ManagerError):
    @property
    def message(self) -> str:
        return 'package already installed'


class PkgConflictsWithInstalled(ManagerError):
    def __init__(self, conflicting_pkgs: Collection[Any]) -> None:
        super().__init__()
        self.conflicting_pkgs = conflicting_pkgs

    @property
    def message(self) -> str:
        return (
            'package folders conflict with installed package'
            + ('s ' if len(self.conflicting_pkgs) > 1 else ' ')
            + ', '.join(f'{c.name} ({c.source}:{c.id})' for c in self.conflicting_pkgs)
        )


class PkgConflictsWithUnreconciled(ManagerError):
    def __init__(self, folders: set[str]) -> None:
        super().__init__()
        self.folders = folders

    @property
    def message(self) -> str:
        folders = ', '.join(f"'{f}'" for f in self.folders)
        return f'package folders conflict with {folders}'


class PkgNonexistent(ManagerError):
    @property
    def message(self) -> str:
        return 'package does not exist'


class PkgFilesMissing(ManagerError):
    def __init__(self, custom_message: str | None = None) -> None:
        super().__init__()
        self._custom_message = custom_message

    @property
    def message(self) -> str:
        return self._custom_message or 'no files are available for download'


class PkgFilesNotMatching(ManagerError):
    def __init__(self, strategy_values: StrategyValues) -> None:
        super().__init__()
        self._strategy_values = strategy_values

    @property
    def message(self) -> str:
        return 'no files found for: ' + '; '.join(
            f'{s}={v!r}' for s, v in asdict(self._strategy_values).items()
        )


class PkgNotInstalled(ManagerError):
    @property
    def message(self) -> str:
        return 'package is not installed'


class PkgSourceInvalid(ManagerError):
    @property
    def message(self) -> str:
        return 'package source is invalid'


class PkgUpToDate(ManagerError):
    def __init__(self, is_pinned: bool) -> None:
        super().__init__()
        self.is_pinned = is_pinned

    @property
    def message(self) -> str:
        return f'package is {"pinned" if self.is_pinned else "up to date"}'


class PkgStrategiesUnsupported(ManagerError):
    def __init__(self, strategies: Collection[Strategy]) -> None:
        super().__init__()
        self.strategies = strategies

    @property
    def message(self) -> str:
        return f'strategies are not valid for source: {", ".join(self.strategies)}'


class InternalError(Result, Exception):
    status: Final = 'error'

    @property
    def message(self) -> str:
        return f'internal error: "{self.args[0]}"'
